"""
Module for calculating window related data. Windows can be indexed relative to
two starting indices.

- Local window index

    - Window index relative to the TimeData is called "local_win"
    - Local window indices always start at 0

- Global window index

    - The global window index is relative to the project reference time
    - The 0 index window begins at the reference time
    - This window indexing is to synchronise data across sites

The global window index is considered the default and sometimes referred to as
the window. Local windows should be explicitly referred to as local_win in
all cases.

The window module includes functionality to do the following:

- Windowing utility functions to calculate window and overlap sizes
- Functions to map windows to samples in TimeData
- Converting a global index array to datetime

Usually with windowing, there is a window size and windows overlap with each
other for a set number of samples. As an illustrative examples, consider a
signal sampled at 10 Hz (dt=0.1 seconds) with 24 samples. This will be windowed
using a window size of 8 samples per window and a 2 sample overlap.

.. plot::

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> fs = 10
    >>> n_samples = 24
    >>> win_size = 8
    >>> olap_size = 2
    >>> times = np.arange(0, n_samples) * (1/fs)

    The first window

    >>> start_win1 = 0
    >>> end_win1 = win_size
    >>> win1_times = times[start_win1:end_win1]

    The second window

    >>> start_win2 = end_win1 - olap_size
    >>> end_win2 = start_win2 + win_size
    >>> win2_times = times[start_win2:end_win2]

    The third window

    >>> start_win3 = end_win2 - olap_size
    >>> end_win3 = start_win3 + win_size
    >>> win3_times = times[start_win3:end_win3]

    The fourth window

    >>> start_win4= end_win3 - olap_size
    >>> end_win4 = start_win4 + win_size
    >>> win4_times = times[start_win4:end_win4]

    Let's look at the actual window times for each window

    >>> win1_times
    array([0. , 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
    >>> win2_times
    array([0.6, 0.7, 0.8, 0.9, 1. , 1.1, 1.2, 1.3])
    >>> win3_times
    array([1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9])
    >>> win4_times
    array([1.8, 1.9, 2. , 2.1, 2.2, 2.3])

    The duration and increments of windows can be calculated using provided
    methods

    >>> from resistics.window import win_duration, inc_duration
    >>> print(win_duration(win_size, fs))
    0:00:00.7
    >>> print(inc_duration(win_size, olap_size, fs))
    0:00:00.6

    Plot the windows to give an illustration of how it works

    >>> plt.plot(win1_times, np.ones_like(win1_times), "bo", label="window1") # doctest: +SKIP
    >>> plt.plot(win2_times, np.ones_like(win2_times)*2, "ro", label="window2") # doctest: +SKIP
    >>> plt.plot(win3_times, np.ones_like(win3_times)*3, "go", label="window3") # doctest: +SKIP
    >>> plt.plot(win4_times, np.ones_like(win4_times)*4, "co", label="window4") # doctest: +SKIP
    >>> plt.xlabel("Time [s]") # doctest: +SKIP
    >>> plt.legend() # doctest: +SKIP
    >>> plt.grid() # doctest: +SKIP
    >>> plt.tight_layout() # doctest: +SKIP
    >>> plt.show() # doctest: +SKIP
"""
from loguru import logger
from pathlib import Path
from typing import Optional, List, Tuple, Dict, Union, Any
from pydantic import PositiveInt
import numpy as np
import pandas as pd

from resistics.errors import ProcessRunError
from resistics.common import History, ResisticsModel, ResisticsData, ResisticsProcess
from resistics.common import ResisticsWriter, Metadata, WriteableMetadata
from resistics.sampling import RSDateTime, RSTimeDelta, HighResDateTime
from resistics.time import ChanMetadata
from resistics.decimate import DecimatedLevelMetadata, DecimatedData


def win_duration(win_size: int, fs: float) -> RSTimeDelta:
    """
    Get the window duration

    Parameters
    ----------
    win_size : int
        Window size in samples
    fs : float
        Sampling frequency Hz

    Returns
    -------
    RSTimeDelta
        Duration

    Examples
    --------
    A few examples with different sampling frequencies and window sizes

    >>> from resistics.window import win_duration
    >>> duration = win_duration(512, 512)
    >>> print(duration)
    0:00:00.998046875
    >>> duration = win_duration(520, 512)
    >>> print(duration)
    0:00:01.013671875
    >>> duration = win_duration(4096, 16_384)
    >>> print(duration)
    0:00:00.24993896484375
    >>> duration = win_duration(200, 0.05)
    >>> print(duration)
    1:06:20
    """
    from resistics.sampling import to_timedelta

    return to_timedelta(1 / fs) * float(win_size - 1)


def inc_duration(win_size: int, olap_size: int, fs: float) -> RSTimeDelta:
    """
    Get the increment between window start times

    If the overlap size = 0, then the time increment between windows is simply
    the window duration. However, when there is an overlap, the increment
    between window start times has to be adjusted by the overlap size

    Parameters
    ----------
    win_size : int
        The window size in samples
    olap_size : int
        The overlap size in samples
    fs : float
        The sample frequency Hz

    Returns
    -------
    RSTimeDelta
        The duration of the window

    Examples
    --------
    >>> from resistics.window import inc_duration
    >>> increment = inc_duration(128, 32, 128)
    >>> print(increment)
    0:00:00.75
    >>> increment = inc_duration(128*3600, 128*60, 128)
    >>> print(increment)
    0:59:00
    """
    from resistics.sampling import to_timedelta

    return to_timedelta(1 / fs) * float(win_size - olap_size)


def win_to_datetime(
    ref_time: RSDateTime, global_win: int, increment: RSTimeDelta
) -> RSDateTime:
    """
    Convert reference window index to start time of window

    Parameters
    ----------
    ref_time : RSDateTime
        Reference time
    global_win : int
        Window index relative to reference time
    increment : RSTimeDelta
        The increment duration

    Returns
    -------
    RSDateTime
        Start time of window

    Examples
    --------
    An example with sampling at 1 Hz, a window size of 100 and an overlap size
    of 25.

    >>> from resistics.sampling import to_datetime
    >>> from resistics.window import inc_duration, win_to_datetime
    >>> ref_time = to_datetime("2021-01-01 00:00:00")
    >>> fs = 1
    >>> win_size = 60
    >>> olap_size = 15
    >>> increment = inc_duration(win_size, olap_size, fs)
    >>> print(increment)
    0:00:45

    The increment is the time increment between the start of time one window and
    the succeeding window.

    >>> print(win_to_datetime(ref_time, 0, increment))
    2021-01-01 00:00:00
    >>> print(win_to_datetime(ref_time, 1, increment))
    2021-01-01 00:00:45
    >>> print(win_to_datetime(ref_time, 2, increment))
    2021-01-01 00:01:30
    >>> print(win_to_datetime(ref_time, 3, increment))
    2021-01-01 00:02:15
    """
    return ref_time + (global_win * increment)


def datetime_to_win(
    ref_time: RSDateTime,
    time: RSDateTime,
    increment: RSTimeDelta,
    method: str = "round",
) -> int:
    """
    Convert a datetime to a global window index

    Parameters
    ----------
    ref_time : RSDateTime
        Reference time
    time : RSDateTime
        Datetime to convert
    increment : RSTimeDelta
        The increment duration
    method : str, optional
        Method for dealing with float results, by default "round"

    Returns
    -------
    int
        The global window index i.e. the window index relative to the reference
        time

    Raises
    ------
    ValueError
        If time < ref_time

    Examples
    --------
    A simple example to show the logic

    >>> from resistics.sampling import to_datetime, to_timedelta
    >>> from resistics.window import datetime_to_win, win_to_datetime, inc_duration
    >>> ref_time = to_datetime("2021-01-01 00:00:00")
    >>> time = to_datetime("2021-01-01 00:01:00")
    >>> increment = to_timedelta(60)
    >>> global_win = datetime_to_win(ref_time, time, increment)
    >>> global_win
    1
    >>> print(win_to_datetime(ref_time, global_win, increment))
    2021-01-01 00:01:00

    A more complex logic with window sizes, overlap sizes and sampling
    frequencies

    >>> fs = 128
    >>> win_size = 256
    >>> olap_size = 64
    >>> ref_time = to_datetime("2021-03-15 00:00:00")
    >>> time = to_datetime("2021-04-17 18:00:00")
    >>> increment = inc_duration(win_size, olap_size, fs)
    >>> print(increment)
    0:00:01.5
    >>> global_win = datetime_to_win(ref_time, time, increment)
    >>> global_win
    1944000
    >>> print(win_to_datetime(ref_time, global_win, increment))
    2021-04-17 18:00:00

    In this scenario, explore the use of rounding

    >>> time = to_datetime("2021-04-17 18:00:00.50")
    >>> global_win = datetime_to_win(ref_time, time, increment, method = "floor")
    >>> global_win
    1944000
    >>> print(win_to_datetime(ref_time, global_win, increment))
    2021-04-17 18:00:00
    >>> global_win = datetime_to_win(ref_time, time, increment, method = "ceil")
    >>> global_win
    1944001
    >>> print(win_to_datetime(ref_time, global_win, increment))
    2021-04-17 18:00:01.5
    >>> global_win = datetime_to_win(ref_time, time, increment, method = "round")
    >>> global_win
    1944000
    >>> print(win_to_datetime(ref_time, global_win, increment))
    2021-04-17 18:00:00

    Another example with a window duration of greater than a day

    >>> fs = 4.8828125e-05
    >>> win_size = 64
    >>> olap_size = 16
    >>> ref_time = to_datetime("1985-07-18 01:00:20")
    >>> time = to_datetime("1985-09-22 23:00:00")
    >>> increment = inc_duration(win_size, olap_size, fs)
    >>> print(increment)
    11 days, 9:04:00
    >>> global_win = datetime_to_win(ref_time, time, increment)
    >>> global_win
    6
    >>> print(win_to_datetime(ref_time, global_win, increment))
    1985-09-24 07:24:20

    This time is greater than the time that was transformed to global window,
    1985-09-22 23:00:00. Try again, this time with the floor option.

    >>> global_win = datetime_to_win(ref_time, time, increment, method="floor")
    >>> global_win
    5
    >>> print(win_to_datetime(ref_time, global_win, increment))
    1985-09-12 22:20:20

    """
    from math import floor, ceil
    from resistics.sampling import to_seconds

    if time < ref_time:
        raise ValueError(f"Time {str(time)} < reference time {str(ref_time)}")

    increment_days_in_seconds, increment_remaining_in_seconds = to_seconds(increment)
    increment_seconds = increment_days_in_seconds + increment_remaining_in_seconds

    delta_days_in_seconds, delta_remaining_in_seconds = to_seconds(time - ref_time)
    delta_total_in_seconds = delta_days_in_seconds + delta_remaining_in_seconds
    n_increments = delta_total_in_seconds / increment_seconds

    if n_increments.is_integer():
        n_increments = int(n_increments)
    elif method == "floor":
        n_increments = int(floor(n_increments))
    elif method == "ceil":
        n_increments = int(ceil(n_increments))
    else:
        n_increments = int(round(n_increments))
    return n_increments


def get_first_and_last_win(
    ref_time: RSDateTime,
    metadata: DecimatedLevelMetadata,
    win_size: int,
    olap_size: int,
) -> Tuple[int, int]:
    """
    Get first and last window for a decimated data level

    .. note::

        For the last window, on initial calculation this may be one or a
        maximum of two windows beyond the last time. The last window is adjusted
        in this function.

        Two windows may occur when the time of the last sample is in the overlap
        of the final two windows.

    Parameters
    ----------
    ref_time : RSDateTime
        The reference time
    metadata : DecimatedLevelMetadata
        Metadata for the decimation level
    win_size : int
        Window size in samples
    olap_size : int
        Overlap size in samples

    Returns
    -------
    Tuple[int, int]
        First and last global windows. This is window indices relative to the
        reference time

    Raises
    ------
    ValueError
        If unable to calculate the last window correctly as this will result in
        an incorrect number of windows

    Examples
    --------
    Get the first and last window for the first decimation level in a decimated
    data instance.

    >>> from resistics.testing import decimated_data_random
    >>> from resistics.sampling import to_datetime
    >>> from resistics.window import get_first_and_last_win, win_to_datetime
    >>> from resistics.window import win_duration, inc_duration
    >>> ref_time = to_datetime("2021-01-01 00:00:00")
    >>> dec_data = decimated_data_random(fs=0.1, first_time="2021-01-01 00:05:10", n_samples=100, factor=10)

    Get the metadata for decimation level 0

    >>> level_metadata = dec_data.metadata.levels_metadata[0]
    >>> level_metadata.summary()
    {
        'fs': 10.0,
        'n_samples': 10000,
        'first_time': '2021-01-01 00:05:10.000000_000000_000000_000000',
        'last_time': '2021-01-01 00:21:49.899999_999999_977300_000000'
    }

    .. note::

        As a point of interest, note how the last time is actually slightly
        incorrect. This is due to machine precision issues described in more
        detail here https://docs.python.org/3/tutorial/floatingpoint.html.
        Whilst there is value in using the high resolution datetime format for
        high sampling rates, there is a tradeoff. Such are the perils of
        floating point arithmetic.

    The next step is to calculate the first and last window, relative to the
    reference time

    >>> win_size = 100
    >>> olap_size = 25
    >>> first_win, last_win = get_first_and_last_win(ref_time, level_metadata, win_size, olap_size)
    >>> print(first_win, last_win)
    42 173

    These window indices can be converted to start times of the windows. The
    last window is checked to make sure it does not extend past the end of the
    time data. First get the window duration and increments.

    >>> duration = win_duration(win_size, level_metadata.fs)
    >>> print(duration)
    0:00:09.9
    >>> increment = inc_duration(win_size, olap_size, level_metadata.fs)
    >>> print(increment)
    0:00:07.5

    Now calculate the times of the windows

    >>> first_win_start_time = win_to_datetime(ref_time, 42, increment)
    >>> last_win_start_time = win_to_datetime(ref_time, 173, increment)
    >>> print(first_win_start_time, last_win_start_time)
    2021-01-01 00:05:15 2021-01-01 00:21:37.5
    >>> print(last_win_start_time + duration)
    2021-01-01 00:21:47.4
    >>> print(level_metadata.last_time)
    2021-01-01 00:21:49.8999999999999773
    >>> level_metadata.last_time > last_win_start_time + increment
    True
    """
    duration = win_duration(win_size, metadata.fs)
    increment = inc_duration(win_size, olap_size, metadata.fs)
    first_win = datetime_to_win(ref_time, metadata.first_time, increment, method="ceil")
    last_win = datetime_to_win(ref_time, metadata.last_time, increment, method="floor")
    # adjust if there is not enough date to complete the last window
    last_win_time = win_to_datetime(ref_time, last_win, increment)
    for attempt in range(2):
        if metadata.last_time >= last_win_time + duration:
            break
        logger.debug(f"Adjusting last window attempt {attempt + 1}")
        last_win -= 1
        last_win_time = win_to_datetime(ref_time, last_win, increment)
    if metadata.last_time < last_win_time + duration:
        raise ValueError("Unable to correctly get the last window")
    return first_win, last_win


def get_win_starts(
    ref_time: RSDateTime,
    win_size: int,
    olap_size: int,
    fs: float,
    n_wins: int,
    index_offset: int,
) -> pd.DatetimeIndex:
    """
    Get window start times

    This is a useful for getting the timestamps for the windows in a dataset

    Parameters
    ----------
    ref_time : RSDateTime
        The reference time
    win_size : int
        The window size
    olap_size : int
        The overlap size
    fs : float
        The sampling frequency
    n_wins : int
        The number of windows
    index_offset : int
        The index offset from the reference time

    Returns
    -------
    pd.DatetimeIndex
        The start times of the windows

    Examples
    --------
    >>> import pandas as pd
    >>> from resistics.sampling import to_datetime
    >>> from resistics.window import get_win_starts
    >>> ref_time = to_datetime("2021-01-01 00:00:00")
    >>> win_size = 100
    >>> olap_size = 25
    >>> fs = 10
    >>> n_wins = 3
    >>> index_offset = 480
    >>> starts = get_win_starts(ref_time, win_size, olap_size, fs, n_wins, index_offset)
    >>> pd.Series(starts)
    0   2021-01-01 01:00:00.000
    1   2021-01-01 01:00:07.500
    2   2021-01-01 01:00:15.000
    dtype: datetime64[ns]
    """
    from resistics.sampling import datetime_array_estimate

    increment = inc_duration(win_size, olap_size, fs)
    first_win_time = win_to_datetime(ref_time, index_offset, increment)
    increment_size = win_size - olap_size
    return datetime_array_estimate(first_win_time, fs / increment_size, n_wins)


def get_win_ends(
    starts: pd.DatetimeIndex,
    win_size: int,
    fs: float,
) -> pd.DatetimeIndex:
    """
    Get window end times

    Parameters
    ----------
    starts : RSDateTime
        The start times of the windows
    win_size : int
        The window size
    fs : float
        The sampling frequency

    Returns
    -------
    pd.DatetimeIndex
        The end times of the windows

    Examples
    --------
    >>> import pandas as pd
    >>> from resistics.sampling import to_datetime
    >>> from resistics.window import get_win_starts, get_win_ends
    >>> ref_time = to_datetime("2021-01-01 00:00:00")
    >>> win_size = 100
    >>> olap_size = 25
    >>> fs = 10
    >>> n_wins = 3
    >>> index_offset = 480
    >>> starts = get_win_starts(ref_time, win_size, olap_size, fs, n_wins, index_offset)
    >>> pd.Series(starts)
    0   2021-01-01 01:00:00.000
    1   2021-01-01 01:00:07.500
    2   2021-01-01 01:00:15.000
    dtype: datetime64[ns]
    >>> ends = get_win_ends(starts, win_size, fs)
    >>> pd.Series(ends)
    0   2021-01-01 01:00:09.900
    1   2021-01-01 01:00:17.400
    2   2021-01-01 01:00:24.900
    dtype: datetime64[ns]
    """
    return starts + pd.Timedelta((win_size - 1) * (1 / fs), "s")


def get_win_table(
    ref_time: RSDateTime,
    metadata: DecimatedLevelMetadata,
    win_size: int,
    olap_size: int,
) -> pd.DataFrame:
    """
    Get a DataFrame with

    Parameters
    ----------
    ref_time : RSDateTime
        Reference
    metadata : DecimatedLevelMetadata
        Metadata for the decimation level
    win_size : int
        The window size
    olap_size : int
        The overlap size

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame with details about each window

    Examples
    --------
    .. plot::
        :width: 90%

        >>> import matplotlib.pyplot as plt
        >>> from resistics.decimate import DecimatedLevelMetadata
        >>> from resistics.sampling import to_datetime, to_timedelta
        >>> from resistics.window import get_win_table
        >>> ref_time = to_datetime("2021-01-01 00:00:00")
        >>> fs = 10
        >>> n_samples = 1000
        >>> first_time = to_datetime("2021-01-01 01:00:00")
        >>> last_time = first_time + to_timedelta((n_samples-1)/fs)
        >>> metadata = DecimatedLevelMetadata(fs=10, n_samples=1000, first_time=first_time, last_time=last_time)
        >>> print(metadata.fs, metadata.first_time, metadata.last_time)
        10.0 2021-01-01 01:00:00 2021-01-01 01:01:39.9
        >>> win_size = 100
        >>> olap_size = 25
        >>> df = get_win_table(ref_time, metadata, win_size, olap_size)
        >>> print(df.to_string())
            global  local  from_sample  to_sample               win_start                 win_end
        0      480      0            0         99 2021-01-01 01:00:00.000 2021-01-01 01:00:09.900
        1      481      1           75        174 2021-01-01 01:00:07.500 2021-01-01 01:00:17.400
        2      482      2          150        249 2021-01-01 01:00:15.000 2021-01-01 01:00:24.900
        3      483      3          225        324 2021-01-01 01:00:22.500 2021-01-01 01:00:32.400
        4      484      4          300        399 2021-01-01 01:00:30.000 2021-01-01 01:00:39.900
        5      485      5          375        474 2021-01-01 01:00:37.500 2021-01-01 01:00:47.400
        6      486      6          450        549 2021-01-01 01:00:45.000 2021-01-01 01:00:54.900
        7      487      7          525        624 2021-01-01 01:00:52.500 2021-01-01 01:01:02.400
        8      488      8          600        699 2021-01-01 01:01:00.000 2021-01-01 01:01:09.900
        9      489      9          675        774 2021-01-01 01:01:07.500 2021-01-01 01:01:17.400
        10     490     10          750        849 2021-01-01 01:01:15.000 2021-01-01 01:01:24.900
        11     491     11          825        924 2021-01-01 01:01:22.500 2021-01-01 01:01:32.400
        12     492     12          900        999 2021-01-01 01:01:30.000 2021-01-01 01:01:39.900

        Plot six windows to illustrate the overlap

        >>> plt.figure(figsize=(8, 3)) # doctest: +SKIP
        >>> for idx, row in df.iterrows():
        ...     color = "red" if idx%2 == 0 else "blue"
        ...     plt.axvspan(row.loc["win_start"], row.loc["win_end"], alpha=0.5, color=color) # doctest: +SKIP
        ...     if idx > 5:
        ...         break
        >>> plt.tight_layout() # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
    """
    from resistics.sampling import to_n_samples, datetime_array_estimate

    increment_size = win_size - olap_size
    increment = inc_duration(win_size, olap_size, metadata.fs)
    fs = metadata.fs
    first_time = metadata.first_time

    first_win, last_win = get_first_and_last_win(
        ref_time, metadata, win_size, olap_size
    )
    first_win_time = win_to_datetime(ref_time, first_win, increment)
    n_wins = last_win - first_win + 1
    local_wins = np.arange(n_wins).astype(int)
    # samples
    first_sample = to_n_samples(first_win_time - first_time, fs, method="round") - 1
    starts = datetime_array_estimate(first_win_time, fs / increment_size, n_wins)
    ends = get_win_ends(starts, win_size, fs)
    df_dict = {
        "global": np.arange(first_win, last_win + 1),
        "local": local_wins,
        "from_sample": first_sample + (local_wins * increment_size),
        "to_sample": first_sample + win_size - 1 + (local_wins * increment_size),
        "win_start": starts,
        "win_end": ends,
    }
    return pd.DataFrame(data=df_dict)


class WindowParameters(ResisticsModel):
    """
    Windowing parameters per decimation level

    Windowing parameters are the window and overlap size for each decimation
    level.

    Parameters
    ----------
    n_levels : int
        The number of decimation levels
    min_n_wins : int
        Minimum number of windows
    win_sizes : List[int]
        The window sizes per decimation level
    olap_sizes : List[int]
        The overlap sizes per decimation level

    Examples
    --------
    Generate decimation and windowing parameters for data sampled at 4096 Hz.
    Note that requesting window sizes or overlap sizes for decimation levels
    that do not exist will raise a ValueError.

    >>> from resistics.decimate import DecimationSetup
    >>> from resistics.window import WindowSetup
    >>> dec_setup = DecimationSetup(n_levels=3, per_level=3)
    >>> dec_params = dec_setup.run(4096)
    >>> dec_params.summary()
    {
        'fs': 4096.0,
        'n_levels': 3,
        'per_level': 3,
        'min_samples': 256,
        'eval_freqs': [
            1024.0,
            724.0773439350246,
            512.0,
            362.0386719675123,
            256.0,
            181.01933598375615,
            128.0,
            90.50966799187808,
            64.0
        ],
        'dec_factors': [1, 2, 8],
        'dec_increments': [1, 2, 4],
        'dec_fs': [4096.0, 2048.0, 512.0]
    }
    >>> win_params = WindowSetup().run(dec_params.n_levels, dec_params.dec_fs)
    >>> win_params.summary()
    {
        'n_levels': 3,
        'min_n_wins': 5,
        'win_sizes': [1024, 512, 256],
        'olap_sizes': [256, 128, 64]
    }
    >>> win_params.get_win_size(0)
    1024
    >>> win_params.get_olap_size(0)
    256
    >>> win_params.get_olap_size(3)
    Traceback (most recent call last):
    ...
    ValueError: Level 3 must be 0 <= level < 3
    """

    n_levels: int
    min_n_wins: int
    win_sizes: List[int]
    olap_sizes: List[int]

    def check_level(self, level: int):
        """Check the decimation level is within range"""
        if level < 0 or level >= self.n_levels:
            raise ValueError(f"Level {level} must be 0 <= level < {self.n_levels}")

    def get_win_size(self, level: int) -> int:
        """Get window size for a decimation level"""
        self.check_level(level)
        return self.win_sizes[level]

    def get_olap_size(self, level: int) -> int:
        """Get overlap size for a decimation level"""
        self.check_level(level)
        return self.olap_sizes[level]


class WindowSetup(ResisticsProcess):
    """
    Setup WindowParameters

    WindowSetup outputs the WindowParameters to use for windowing decimated
    time data.

    Window parameters are simply the window and overlap sizes for each
    decimation level.

    Parameters
    ----------
    min_size : int, optional
        Minimum window size, by default 256
    min_olap : int, optional
        Minimum overlap size, by default 64
    win_factor : int, optional
        Window factor, by default 4. Window sizes are calculated by sampling
        frequency / 4 to ensure sufficient frequency resolution. If the
        sampling frequency is small, window size will be adjusted to
        min_size
    olap_proportion : float, optional
        The proportion of the window size to use as the overlap, by default
        0.25. For example, for a window size of 128, the overlap would be
        0.25 * 128 = 32
    min_n_wins : int, optional
        The minimum number of windows needed in a decimation level, by
        default 5
    win_sizes : Optional[List[int]], optional
        Explicit define window sizes, by default None. Must have the same
        length as number of decimation levels
    olap_sizes : Optional[List[int]], optional
        Explicitly define overlap sizes, by default None. Must have the same
        length as number of decimation levels

    Examples
    --------
    Generate decimation and windowing parameters for data sampled at 0.05 Hz or
    20 seconds sampling period

    >>> from resistics.decimate import DecimationSetup
    >>> from resistics.window import WindowSetup
    >>> dec_params = DecimationSetup(n_levels=3, per_level=3).run(0.05)
    >>> dec_params.summary()
    {
        'fs': 0.05,
        'n_levels': 3,
        'per_level': 3,
        'min_samples': 256,
        'eval_freqs': [
            0.0125,
            0.008838834764831844,
            0.00625,
            0.004419417382415922,
            0.003125,
            0.002209708691207961,
            0.0015625,
            0.0011048543456039805,
            0.00078125
        ],
        'dec_factors': [1, 2, 8],
        'dec_increments': [1, 2, 4],
        'dec_fs': [0.05, 0.025, 0.00625]
    }
    >>> win_params = WindowSetup().run(dec_params.n_levels, dec_params.dec_fs)
    >>> win_params.summary()
    {
        'n_levels': 3,
        'min_n_wins': 5,
        'win_sizes': [256, 256, 256],
        'olap_sizes': [64, 64, 64]
    }

    Window parameters can also be explicitly defined

    >>> from resistics.decimate import DecimationSetup
    >>> from resistics.window import WindowSetup
    >>> dec_setup = DecimationSetup(n_levels=3, per_level=3)
    >>> dec_params = dec_setup.run(0.05)
    >>> win_setup = WindowSetup(win_sizes=[1000, 578, 104])
    >>> win_params = win_setup.run(dec_params.n_levels, dec_params.dec_fs)
    >>> win_params.summary()
    {
        'n_levels': 3,
        'min_n_wins': 5,
        'win_sizes': [1000, 578, 104],
        'olap_sizes': [250, 144, 64]
    }
    """

    min_size: int = 256
    min_olap: int = 64
    win_factor: int = 4
    olap_proportion: float = 0.25
    min_n_wins: int = 5
    win_sizes: Optional[List[int]] = None
    olap_sizes: Optional[List[int]] = None

    def run(self, n_levels: int, dec_fs: List[float]) -> WindowParameters:
        """
        Calculate window and overlap sizes for each decimation level based on
        decimation level sampling frequency and minimum allowable parameters

        The window and overlap sizes (number of samples) are calculated based in
        the following way:

        - window size = frequency at decimation level / window factor
        - overlap size = window size * overlap proportion

        This is to ensure good frequency resolution at high frequencies. At low
        sampling frequencies, this would result in very small window sizes,
        therefore, there a minimum allowable sizes for both windows and overlap
        defined by min_size and min_olap in the initialiser. If window sizes
        or overlaps size are calculated below these respecitively, they will be
        set to the minimum values.

        Parameters
        ----------
        n_levels : int
            The number of decimation levels
        dec_fs : List[float]
            The sampling frequencies for each decimation level

        Returns
        -------
        WindowParameters
            The window parameters, the window sizes and overlaps for each
            decimation level

        Raises
        ------
        ValueError
            If the number of windows does not match the number of levels
        ValueError
            If the number of overlaps does not match the number of levels
        """
        if self.win_sizes is None:
            win_sizes = self._get_win_sizes(n_levels, dec_fs)
        else:
            win_sizes = list(self.win_sizes)

        if len(win_sizes) < n_levels:
            raise ValueError(f"Num. windows {len(win_sizes)} < n_levels {n_levels}")
        if len(win_sizes) > n_levels:
            # this may happen with user input windows
            # but decimated data has fewer levels
            win_sizes = win_sizes[:n_levels]

        if self.olap_sizes is None:
            olap_sizes = self._get_olap_sizes(win_sizes)
        else:
            olap_sizes = self.olap_sizes

        if len(olap_sizes) < n_levels:
            raise ValueError(f"Num. overlaps {len(olap_sizes)} < n_levels {n_levels}")
        if len(olap_sizes) > n_levels:
            # this may happen with user input windows
            # but decimated data has fewer levels
            olap_sizes = olap_sizes[:n_levels]

        return WindowParameters(
            n_levels=n_levels,
            min_n_wins=self.min_n_wins,
            win_sizes=win_sizes,
            olap_sizes=olap_sizes,
        )

    def _get_win_sizes(self, n_levels: int, dec_fs: List[float]) -> List[int]:
        """
        Get the window sizes

        Parameters
        ----------
        n_levels : int
            The number of decimation levels
        dec_fs : List[float]
            The sampling frequencies for each decimation level

        Returns
        -------
        List[int]
            Window sizes
        """
        win_sizes = []
        for ilevel in range(n_levels):
            win_size = dec_fs[ilevel] // self.win_factor
            if win_size < self.min_size:
                win_size = self.min_size
            win_sizes.append(int(win_size))
        return win_sizes

    def _get_olap_sizes(self, win_sizes: List[int]) -> List[int]:
        """
        Get overlap sizes

        Parameters
        ----------
        win_sizes : List[int]
            The window sizes

        Returns
        -------
        List[int]
            The overlap sizes
        """
        olap_sizes = []
        for win_size in win_sizes:
            olap_size = int(win_size * self.olap_proportion)
            if olap_size < self.min_olap:
                olap_size = self.min_olap
            olap_sizes.append(olap_size)
        return olap_sizes


class WindowedLevelMetadata(Metadata):
    """Metadata for a windowed level"""

    fs: float
    """The sampling frequency for the decimation level"""
    n_wins: int
    """The number of windows"""
    win_size: PositiveInt
    """The window size in samples"""
    olap_size: PositiveInt
    """The overlap size in samples"""
    index_offset: int
    """The global window offset for local window 0"""

    @property
    def dt(self):
        return 1 / self.fs


class WindowedMetadata(WriteableMetadata):
    """Metadata for windowed data"""

    fs: List[float]
    chans: List[str]
    n_chans: Optional[int] = None
    n_levels: int
    first_time: HighResDateTime
    last_time: HighResDateTime
    system: str = ""
    serial: str = ""
    wgs84_latitude: float = -999.0
    wgs84_longitude: float = -999.0
    easting: float = -999.0
    northing: float = -999.0
    elevation: float = -999.0
    chans_metadata: Dict[str, ChanMetadata]
    levels_metadata: List[WindowedLevelMetadata]
    ref_time: HighResDateTime
    history: History = History()

    class Config:

        extra = "ignore"


class WindowedData(ResisticsData):
    """
    Windows of a DecimatedData object

    The windowed data is stored in a dictionary attribute named data. This is
    a dictionary with an entry for each decimation level. The shape for a single
    decimation level is as follows:

    n_wins x n_chans x n_samples
    """

    def __init__(
        self,
        metadata: WindowedMetadata,
        data: Dict[int, np.ndarray],
    ):
        """
        Initialise the WindowedData

        Parameters
        ----------
        metadata : WindowedDataMetadata
            The metadata for the windowed data
        data : Dict[int, WindowedTimeData]
            The windowed data
        """
        logger.debug(f"Creating WindowedData with data type {data[0].dtype}")
        self.metadata = metadata
        self.data = data

    def get_level(self, level: int) -> np.ndarray:
        """
        Get windows for a decimation level

        Parameters
        ----------
        level : int
            The decimation level

        Returns
        -------
        np.ndarray
            The window array

        Raises
        ------
        ValueError
            If decimation level is not within range
        """
        if level >= self.metadata.n_levels:
            raise ValueError(f"Level {level} not <= max {self.metadata.n_levels - 1}")
        return self.data[level]

    def get_local(self, level: int, local_win: int) -> np.ndarray:
        """
        Get window using local index

        Parameters
        ----------
        level : int
            The decimation level
        local_win : int
            Local window index

        Returns
        -------
        np.ndarray
            Window data

        Raises
        ------
        ValueError
            If local window index is out of range
        """
        n_wins = self.metadata.levels_metadata[level].n_wins
        if local_win < 0 or local_win >= n_wins:
            raise ValueError(f"Local window {local_win} not 0 <= local_win < {n_wins}")
        return self.get_level(level)[local_win]

    def get_global(self, level: int, global_win: int) -> np.ndarray:
        """
        Get window using global index

        Parameters
        ----------
        level : int
            The decimation level
        global_win : int
            Global window index

        Returns
        -------
        np.ndarray
            Window data
        """
        index_offset = self.metadata.levels_metadata[level].index_offset
        return self.get_local(level, global_win + index_offset)

    def get_chan(self, level: int, chan: str) -> np.ndarray:
        """
        Get all the windows for a channel

        Parameters
        ----------
        level : int
            The decimation level
        chan : str
            The channel

        Returns
        -------
        np.ndarray
            The data for the channels

        Raises
        ------
        ChannelNotFoundError
            If the channel is not found in the data
        """
        from resistics.errors import ChannelNotFoundError

        if chan not in self.metadata.chans:
            raise ChannelNotFoundError(chan, self.metadata.chans)
        idx = self.metadata.chans.index(chan)
        return self.get_level(level)[..., idx, :]

    def to_string(self) -> str:
        """Class information as a string"""
        return self.metadata.to_string()


class Windower(ResisticsProcess):
    """
    Windows DecimatedData

    This is the primary window making process for resistics and should be used
    when alignment of windows with a site or across sites is required.

    This method uses numpy striding to produce window views into the decimated
    data.

    See Also
    --------
    WindowerTarget : A windower to make a target number of windows

    Examples
    --------
    The Windower windows a DecimatedData object given a reference time and some
    window parameters.

    There's quite a few imports needed for this example. Begin by doing the
    imports, defining a reference time and generating random decimated data.

    >>> from resistics.sampling import to_datetime
    >>> from resistics.testing import decimated_data_linear
    >>> from resistics.window import WindowSetup, Windower
    >>> dec_data = decimated_data_linear(fs=128)
    >>> ref_time = dec_data.metadata.first_time
    >>> print(dec_data.to_string())
    <class 'resistics.decimate.DecimatedData'>
               fs        dt  n_samples           first_time                        last_time
    level
    0      2048.0  0.000488      16384  2021-01-01 00:00:00  2021-01-01 00:00:07.99951171875
    1       512.0  0.001953       4096  2021-01-01 00:00:00    2021-01-01 00:00:07.998046875
    2       128.0  0.007812       1024  2021-01-01 00:00:00      2021-01-01 00:00:07.9921875

    Next, initialise the window parameters. For this example, use small windows,
    which will make inspecting them easier.

    >>> win_params = WindowSetup(win_sizes=[16,16,16], min_olap=4).run(dec_data.metadata.n_levels, dec_data.metadata.fs)
    >>> win_params.summary()
    {
        'n_levels': 3,
        'min_n_wins': 5,
        'win_sizes': [16, 16, 16],
        'olap_sizes': [4, 4, 4]
    }

    Perform the windowing. This actually creates views into the decimated data
    using the numpy.lib.stride_tricks.sliding_window_view function. The shape
    for a data array at a decimation level is: n_wins x n_chans x win_size. The
    information about each level is also in the levels_metadata attribute of
    WindowedMetadata.

    >>> win_data = Windower().run(ref_time, win_params, dec_data)
    >>> win_data.data[0].shape
    (1365, 2, 16)
    >>> for level_metadata in win_data.metadata.levels_metadata:
    ...     level_metadata.summary()
    {
        'fs': 2048.0,
        'n_wins': 1365,
        'win_size': 16,
        'olap_size': 4,
        'index_offset': 0
    }
    {
        'fs': 512.0,
        'n_wins': 341,
        'win_size': 16,
        'olap_size': 4,
        'index_offset': 0
    }
    {
        'fs': 128.0,
        'n_wins': 85,
        'win_size': 16,
        'olap_size': 4,
        'index_offset': 0
    }

    Let's look at an example of data from the first decimation level for the
    first channel. This is simply a linear set of data ranging from 0...16_383.

    >>> dec_data.data[0][0]
    array([    0,     1,     2, ..., 16381, 16382, 16383])

    Inspecting the first few windows shows they are as expected including the
    overlap.

    >>> win_data.data[0][0, 0]
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15])
    >>> win_data.data[0][1, 0]
    array([12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27])
    >>> win_data.data[0][2, 0]
    array([24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39])
    """

    def run(
        self,
        ref_time: RSDateTime,
        win_params: WindowParameters,
        dec_data: DecimatedData,
    ) -> WindowedData:
        """
        Perform windowing of DecimatedData

        Parameters
        ----------
        ref_time : RSDateTime
            The reference time
        win_params : WindowParameters
            The window parameters
        dec_data : DecimatedData
            The decimated data

        Returns
        -------
        WindowedData
            Windows for decimated data

        Raises
        ------
        ProcessRunError
            If the number of windows calculated in the window table does not
            match the size of the array views
        """
        metadata_dict = dec_data.metadata.dict()
        data = {}
        win_levels_metadata = []
        messages = []
        for ilevel in range(0, dec_data.metadata.n_levels):
            logger.info(f"Windowing decimation level {ilevel}")
            win_size = win_params.get_win_size(ilevel)
            olap_size = win_params.get_olap_size(ilevel)
            level_metadata = dec_data.metadata.levels_metadata[ilevel]
            win_table = get_win_table(ref_time, level_metadata, win_size, olap_size)
            n_wins = len(win_table.index)
            logger.info(f"{n_wins} windows, size {win_size}, overlap {olap_size}")
            messages.append(f"Level {ilevel}, generated {n_wins} windows")
            messages.append(f"Window size {win_size}, olap_size {olap_size}")

            if n_wins < win_params.min_n_wins:
                logger.debug(f"Number windows {n_wins} < min. {win_params.min_n_wins}")
                messages.append(f"Num. windows {n_wins} < min. {win_params.min_n_wins}")
                messages.append(f"Level {ilevel} incomplete, terminating windowing")
                break

            win_level_data = self._get_level_data(
                dec_data.get_level(ilevel),
                win_table,
                dec_data.metadata.n_chans,
                win_size,
                olap_size,
            )
            if win_level_data.shape[0] != n_wins:
                raise ProcessRunError(
                    self.name,
                    f"Num. windows mismatch {win_level_data.shape[0]} != {n_wins}",
                )
            win_level_metadata = self._get_level_metadata(
                level_metadata,
                win_table,
                win_size,
                olap_size,
            )
            data[ilevel] = win_level_data
            win_levels_metadata.append(win_level_metadata)
        metadata_dict["ref_time"] = ref_time
        metadata = self._get_metadata(metadata_dict, win_levels_metadata)
        metadata.history.add_record(self._get_record(messages))
        logger.info("Windowing completed")
        return WindowedData(metadata, data)

    def _get_level_data(
        self,
        data: np.ndarray,
        win_table: pd.DataFrame,
        n_chans: int,
        win_size: int,
        olap_size: int,
    ) -> np.ndarray:
        """
        Get window data for a decimation level

        Parameters
        ----------
        data : np.ndarray
            The decimated time data for the level
        win_table : pd.DataFrame
            The window table
        n_chans : int
            The number of channels
        win_size : int
            The window size
        olap_size : int
            The overlap size

        Returns
        -------
        np.ndarray
            Sliding window views in an array for the decimation level
        """
        from numpy.lib.stride_tricks import sliding_window_view

        from_sample = win_table.loc[0, "from_sample"]
        increment_size = win_size - olap_size
        view = np.squeeze(
            sliding_window_view(data, window_shape=(n_chans, win_size), writeable=True)
        )
        return view[from_sample::increment_size]

    def _get_level_metadata(
        self,
        level_metadata: DecimatedLevelMetadata,
        win_table: pd.DataFrame,
        win_size: int,
        olap_size: int,
    ) -> WindowedLevelMetadata:
        """Get the windowed metadata for a decimation level"""
        offset = (win_table["global"] - win_table["local"]).unique()
        if len(offset) != 1:
            raise ValueError("Malformed window table, varying local to global offset")
        return WindowedLevelMetadata(
            fs=level_metadata.fs,
            n_wins=len(win_table.index),
            win_size=win_size,
            olap_size=olap_size,
            index_offset=offset[0],
        )

    def _get_metadata(
        self,
        metadata_dict: Dict[str, Any],
        levels_metadata: List[WindowedLevelMetadata],
    ) -> WindowedMetadata:
        """Get the metadata for the windowed data"""
        metadata_dict.pop("file_info")
        metadata_dict["n_levels"] = len(levels_metadata)
        metadata_dict["levels_metadata"] = levels_metadata
        return WindowedMetadata(**metadata_dict)


class WindowerTarget(Windower):
    """
    Windower that selects window sizes to meet a target number of windows

    The minimum window size in window parameters will be respected even if the
    generated number of windows is below the target. This is to avoid situations
    where excessively small windows sizes are selected.

    .. warning::

        This process is primarily useful for quick processing of a single
        measurement and should not be used when any alignment of windows is
        required within a site or across sites.

    Parameters
    ----------
    target : int
        The target number of windows for each decimation level
    olap_proportion : float
        The overlap proportion of the window size

    See Also
    --------
    Windower : The window making process to use when alignment is required
    """

    target: int = 1000
    min_size: int = 64
    olap_proportion: float = 0.25

    def run(
        self,
        ref_time: RSDateTime,
        win_params: WindowParameters,
        dec_data: DecimatedData,
    ) -> WindowedData:
        metadata_dict = dec_data.metadata.dict()
        data = {}
        win_levels_metadata = []
        messages = []
        for ilevel in range(0, dec_data.metadata.n_levels):
            logger.info(f"Windowing decimation level {ilevel}")
            level_metadata = dec_data.metadata.levels_metadata[ilevel]
            win_size = self._get_win_size(level_metadata)
            olap_size = int(np.floor(self.olap_proportion * win_size))
            win_table = get_win_table(ref_time, level_metadata, win_size, olap_size)
            n_wins = len(win_table.index)
            logger.info(f"{n_wins} windows, size {win_size}, overlap {olap_size}")
            messages.append(f"Level {ilevel}, generated {n_wins} windows")
            messages.append(f"Window size {win_size}, olap_size {olap_size}")

            if n_wins < win_params.min_n_wins:
                logger.debug(f"Number windows {n_wins} < min. {win_params.min_n_wins}")
                messages.append(f"Num. windows {n_wins} < min. {win_params.min_n_wins}")
                messages.append(f"Level {ilevel} incomplete, terminating windowing")
                break

            win_level_data = self._get_level_data(
                dec_data.get_level(ilevel),
                win_table,
                dec_data.metadata.n_chans,
                win_size,
                olap_size,
            )
            win_level_metadata = self._get_level_metadata(
                level_metadata,
                win_table,
                win_size,
                olap_size,
            )
            data[ilevel] = win_level_data
            win_levels_metadata.append(win_level_metadata)
        metadata_dict["ref_time"] = metadata_dict["first_time"]
        metadata = self._get_metadata(metadata_dict, win_levels_metadata)
        metadata.history.add_record(self._get_record(messages))
        logger.info("Windowing completed")
        return WindowedData(metadata, data)

    def _get_win_size(self, level_metadata: DecimatedLevelMetadata) -> int:
        r"""
        Get window size that gives close to the target number of windows

        Windows increment by (window size - overlap size), therefore the
        follwing equation is solved,

        .. math::

            n_{samples} / ((1 - n_{overlap})*n_{window}) = target

        Rearrangning, get,

        .. math::

            n_{window} = n_{samples} / ((1 - n_{overlap})*target)

        Parameters
        ----------
        level_metadata : DecimatedLevelMetadata
            The metadata for the decimation level

        Returns
        -------
        int
            The window size
        """
        win_size = level_metadata.n_samples / ((1 - self.olap_proportion) * self.target)
        win_size = int(np.floor(win_size))
        if win_size < self.min_size:
            return self.min_size
        return win_size


class WindowedDataWriter(ResisticsWriter):
    """Writer of resistics windowed data"""

    def run(self, dir_path: Path, win_data: WindowedData) -> None:
        """
        Write out WindowedData

        Parameters
        ----------
        dir_path : Path
            The directory path to write to
        win_data : WindowedData
            Windowed data to write out

        Raises
        ------
        WriteError
            If unable to write to the directory
        """
        from resistics.errors import WriteError

        if not self._check_dir(dir_path):
            raise WriteError(dir_path, "Unable to write to directory, check logs")
        logger.info(f"Writing windowed data to {dir_path}")
        metadata_path = dir_path / "metadata.json"
        data_path = dir_path / "data"
        np.savez_compressed(data_path, **{str(x): y for x, y in win_data.data.items()})
        metadata = win_data.metadata.copy()
        metadata.history.add_record(self._get_record(dir_path, type(win_data)))
        metadata.write(metadata_path)


class WindowedDataReader(ResisticsProcess):
    """Reader of resistics windowed data"""

    def run(
        self, dir_path: Path, metadata_only: bool = False
    ) -> Union[WindowedMetadata, WindowedData]:
        """
        Read WindowedData

        Parameters
        ----------
        dir_path : Path
            The directory path to read from
        metadata_only : bool, optional
            Flag for getting metadata only, by default False

        Returns
        -------
        Union[WindowedMetadata, WindowedData]
            The WindowedData or WindowedMetadata if metadata_only is True

        Raises
        ------
        ReadError
            If the directory does not exist
        """
        from resistics.errors import ReadError

        if not dir_path.exists():
            raise ReadError(dir_path, "Directory does not exist")
        logger.info(f"Reading windowed data from {dir_path}")
        metadata_path = dir_path / "metadata.json"
        metadata = WindowedMetadata.parse_file(metadata_path)
        if metadata_only:
            return metadata
        data_path = dir_path / "data.npz"
        npz_file = np.load(data_path)
        data = {int(level): npz_file[level] for level in npz_file.files}
        messages = [f"Windowed data read from {dir_path}"]
        metadata.history.add_record(self._get_record(messages))
        return WindowedData(metadata, data)



"""
Module for producing testing data for resistics

This includes testing data for:

- Record
- History
- TimeMetadata
- TimeData
- DecimatedData
- SpectraData
"""

from typing import List, Dict, Optional, Type, Union
import numpy as np
import pandas as pd

from resistics.common import Record, History, get_record
from resistics.time import get_time_metadata, TimeMetadata, TimeData
from resistics.decimate import DecimatedMetadata, DecimatedData
from resistics.spectra import SpectraLevelMetadata, SpectraMetadata, SpectraData
from resistics.gather import SiteCombinedMetadata
from resistics.transfunc import Component, ImpedanceTensor
from resistics.regression import RegressionInputMetadata, Solution

DEFAULT_TIME_DATA_DTYPE = np.float32


def record_example1() -> Record:
    """Get an example Record"""
    from resistics.common import get_record

    return get_record(
        creator={"name": "example1", "a": 5, "b": -7.0},
        messages=["Message 1", "Message 2"],
    )


def record_example2() -> Record:
    """Get an example Record"""
    from resistics.common import get_record

    return get_record(
        creator={"name": "example2", "a": "parzen", "b": -21},
        messages=["Message 5", "Message 6"],
    )


def history_example() -> History:
    """Get a History example"""
    from resistics.common import History

    return History(records=[record_example1(), record_example2()])


def time_metadata_1chan(
    fs: float = 10, first_time: str = "2021-01-01 00:00:00", n_samples: int = 11
) -> TimeMetadata:
    """
    Get TimeMetadata for a single channel, "chan1"

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The first time, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 11

    Returns
    -------
    TimeMetadata
        TimeMetadata
    """
    first_time = pd.to_datetime(first_time)
    time_dict = {
        "chans": ["chan1"],
        "fs": fs,
        "n_samples": n_samples,
        "n_chans": 1,
        "first_time": first_time,
        "last_time": first_time + pd.Timedelta(1 / fs, "s") * (n_samples - 1),
    }
    chans_dict = {
        "chan1": {
            "name": "chan1",
            "chan_type": "electric",
            "data_files": "example1.ascii",
        },
    }
    return get_time_metadata(time_dict, chans_dict)


def time_metadata_2chan(
    fs: float = 10, first_time: str = "2021-01-01 00:00:00", n_samples: int = 11
) -> TimeMetadata:
    """
    Get a TimeMetadata instance with two channels, "chan1" and "chan2"

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The first time, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 11

    Returns
    -------
    TimeMetadata
        TimeMetadata
    """
    first_time = pd.to_datetime(first_time)
    time_dict = {
        "chans": ["chan1", "chan2"],
        "fs": fs,
        "n_samples": n_samples,
        "n_chans": 2,
        "first_time": first_time,
        "last_time": first_time + pd.Timedelta(1 / fs, "s") * (n_samples - 1),
    }
    chans_dict = {
        "chan1": {
            "name": "chan1",
            "chan_type": "electric",
            "data_files": "example1.ascii",
        },
        "chan2": {
            "name": "chan2",
            "chan_type": "magnetic",
            "data_files": "example2.ascii",
            "sensor": "MFS",
        },
    }
    return get_time_metadata(time_dict, chans_dict)


def time_metadata_mt(
    fs: float = 10, first_time: str = "2020-01-01 00:00:00", n_samples: int = 11
) -> TimeMetadata:
    """
    Get a magnetotelluric time metadata with four channels "Ex", "Ey", "Hx", "Hy"

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The first time, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 11

    Returns
    -------
    TimeMetadata
        TimeMetadata
    """
    first_time = pd.to_datetime(first_time)
    time_dict = {
        "chans": ["Ex", "Ey", "Hx", "Hy"],
        "fs": fs,
        "dt": 1 / fs,
        "n_chans": 4,
        "n_samples": n_samples,
        "first_time": first_time,
        "last_time": first_time + pd.Timedelta(1 / fs, "s") * (n_samples - 1),
    }
    chans_dict = {
        "Ex": {"name": "Ex", "data_files": "Ex.ascii"},
        "Ey": {"name": "Ey", "data_files": "Ex.ascii"},
        "Hx": {"name": "Hx", "data_files": "Ex.ascii"},
        "Hy": {"name": "Hy", "data_files": "Ex.ascii"},
    }
    return get_time_metadata(time_dict, chans_dict)


def time_data_ones(
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    n_samples: int = 10,
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    TimeData with all ones

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 10
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        The TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    metadata = time_metadata_mt(fs, first_time, n_samples)
    data = np.ones(shape=(len(metadata.chans), n_samples), dtype=dtype)
    creator = {
        "name": "time_data_ones",
        "fs": fs,
        "first_time": first_time,
        "n_samples": n_samples,
    }
    messages = ["Generated time data with fixed values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_simple(
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    Time data with 16 samples

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2020-01-01 00:00:00"
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        The TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    data = np.array(
        [
            [1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1],
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4, 5, 6, 7],
            [1, 2, 3, 5, 1, 2, 3, 4, 2, 6, 7, 6, 5, 4, 3, 2],
            [2, 0, 0, 1, 2, 3, 0, 0, 0, 0, 0, 1, 3, 3, 3, 3],
        ],
        dtype=dtype,
    )
    n_samples = data.shape[1]
    metadata = time_metadata_mt(fs, first_time, n_samples)
    creator = {"name": "time_data_simple", "fs": fs, "first_time": first_time}
    messages = ["Generated time data with simple values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_with_nans(
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    TimeData with 16 samples and some nan values

    Parameters
    ----------
    fs : float, optional
        Sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2020-01-01 00:00:00"
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        The TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    data = np.array(
        [
            [1, 1, 1, 0, np.nan, 0, 1, 1, 1, np.nan, 0, 0, 0, 0, 1, 1],
            [1, 2, np.nan, np.nan, 5, 6, np.nan, 8, 9, 1, 2, 3, 4, 5, 6, 7],
            [np.nan, 2, 3, 5, 1, 2, 3, 4, 2, 6, 7, np.nan, np.nan, 4, 3, 2],
            [2, 0, 0, 1, 2, 3, np.nan, np.nan, np.nan, 0, 0, 1, 3, 3, 3, 3],
        ],
        dtype=dtype,
    )
    n_samples = data.shape[1]
    metadata = time_metadata_mt(fs, first_time, n_samples)
    creator = {"name": "time_data_with_nans", "fs": fs, "first_time": first_time}
    messages = ["Generated time data with some nan values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_linear(
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    n_samples: int = 10,
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    Get TimeData with linear data

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        Time of first sample, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 10
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        TimeData with linear values
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    metadata = time_metadata_mt(fs, first_time, n_samples)
    data = np.empty(shape=(metadata.n_chans, n_samples), dtype=dtype)
    for idx in range(metadata.n_chans):
        data[idx, :] = np.arange(n_samples)
    creator = {
        "name": "time_data_linear",
        "fs": fs,
        "first_time": first_time,
        "n_samples": n_samples,
    }
    messages = ["Generated time data with linear values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_random(
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    n_samples: int = 10,
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    TimeData with random values and specifiable number of samples

    Parameters
    ----------
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        Time of first sample, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 10
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        The TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    metadata = time_metadata_mt(fs, first_time, n_samples)
    data = np.random.normal(0, 3, size=(metadata.n_chans, n_samples)).astype(dtype)
    creator = {
        "name": "time_data_random",
        "fs": fs,
        "first_time": first_time,
        "n_samples": n_samples,
    }
    messages = ["Generated time data with random values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_periodic(
    frequencies: List[float],
    fs: float = 50,
    first_time: str = "2020-01-01 00:00:00",
    n_samples: int = 100,
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    Get period TimeData

    Parameters
    ----------
    frequencies : List[float]
        Frequencies to include
    fs : float, optional
        Sampling frequency, by default 50
    first_time : str, optional
        The first time, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 100
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        Periodic TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    metadata = time_metadata_1chan(fs, first_time, n_samples)
    times = np.arange(0, n_samples) * (1 / fs)
    data = np.zeros(shape=(1, n_samples), dtype=dtype)
    for freq in frequencies:
        data += np.sin(times * 2 * np.pi * freq)
    creator = {
        "name": "time_data_periodic",
        "frequencies": frequencies,
        "fs": fs,
        "first_time": first_time,
        "n_samples": n_samples,
    }
    messages = ["Generated time data with periodic values"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def time_data_with_offset(
    offset=0.05,
    fs: float = 10,
    first_time: str = "2020-01-01 00:00:00",
    n_samples: int = 11,
    dtype: Optional[Type] = None,
) -> TimeData:
    """
    Get TimeData with an offset on the sampling

    Parameters
    ----------
    offset : float, optional
        The offset on the sampling in seconds, by default 0.05
    fs : float, optional
        The sampling frequency, by default 10
    first_time : str, optional
        The first time of the TimeData, by default "2020-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 11
    dtype : Optional[Type], optional
        The data type for the values, by default None

    Returns
    -------
    TimeData
        The TimeData
    """
    if dtype is None:
        dtype = DEFAULT_TIME_DATA_DTYPE
    first_time = (pd.to_datetime(first_time) + pd.Timedelta(offset, "s")).isoformat()
    metadata = time_metadata_1chan(fs, first_time, n_samples)
    data = np.arange(0, n_samples).reshape(1, n_samples)
    creator = {
        "name": "time_data_with_offset",
        "offset": offset,
        "fs": fs,
        "first_time": first_time,
        "n_samples": n_samples,
    }
    messages = ["Generated time data with an offset"]
    record = get_record(creator, messages)
    metadata.history.add_record(record)
    return TimeData(metadata, data)


def decimated_metadata(
    fs: float = 0.25,
    first_time: str = "2021-01-01 00:00:00",
    n_samples: int = 1024,
    n_levels: int = 3,
    factor: int = 4,
) -> DecimatedMetadata:
    """
    Get example decimated metadata

    The final level has n_samples. The number of samples for all other
    levels is calculated using a decimation factor of 4.

    Similarly for the sampling frequencies, the final level is assumed to have
    a sample frequency of fs and all other levels sampling frequencies are
    calculated from there.

    Parameters
    ----------
    fs : float, optional
        The sampling frequency of the last level, by default 0.25
    first_time : str, optional
        The time of the first sample, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 1024
    n_levels : int, optional
        The number of decimation levels, by default 3
    factor : int, optional
        The decimation factor for each level, by default 4

    Returns
    -------
    DecimatedMetadata
        DecimatedMetadata
    """
    from resistics.decimate import DecimatedLevelMetadata
    from resistics.sampling import to_datetime, to_timedelta

    levels_metadata = []
    for ilevel in range(n_levels):
        level_factor = np.power(factor, (n_levels - 1 - ilevel))
        level_n_samples = n_samples * level_factor
        level_fs = fs * level_factor
        last_time = to_datetime(first_time) + to_timedelta(
            (level_n_samples - 1) * 1 / level_fs
        )
        metadata = DecimatedLevelMetadata(
            fs=level_fs,
            n_samples=level_n_samples,
            first_time=first_time,
            last_time=last_time,
        )
        levels_metadata.append(metadata)
    fs = levels_metadata[0].fs
    n_samples = levels_metadata[0].n_samples
    time_metadata = time_metadata_2chan(
        fs=fs, first_time=first_time, n_samples=n_samples
    )
    metadata_dict = time_metadata.dict()
    metadata_dict["fs"] = [x.fs for x in levels_metadata]
    metadata_dict["n_levels"] = n_levels
    metadata_dict["levels_metadata"] = levels_metadata
    return DecimatedMetadata(**metadata_dict)


def decimated_data_random(
    fs: float = 0.25,
    first_time: str = "2021-01-01 00:00:00",
    n_samples: int = 1024,
    n_levels: int = 3,
    factor: int = 4,
) -> DecimatedData:
    """
    Get random decimated data

    Parameters
    ----------
    fs : float, optional
        Sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 1024
    n_levels : int, optional
        The number of levels, by default 3
    factor : int, optional
        The decimation factor for each level, by default 4

    Returns
    -------
    DecimatedData
        The decimated data
    """
    metadata = decimated_metadata(
        fs, first_time, n_samples=n_samples, n_levels=n_levels, factor=factor
    )
    data = {}
    for ilevel in range(metadata.n_levels):
        level_samples = metadata.levels_metadata[ilevel].n_samples
        data[ilevel] = np.random.normal(0, 3, size=(metadata.n_chans, level_samples))
    creator = {
        "name": "decimated_data_random",
        "fs": fs,
        "first_time": first_time,
        "n_levels": n_levels,
    }
    record = get_record(creator, "Generated random decimated data")
    metadata.history.add_record(record)
    return DecimatedData(metadata, data)


def decimated_data_linear(
    fs: float = 0.25,
    first_time: str = "2021-01-01 00:00:00",
    n_samples: int = 1024,
    n_levels: int = 3,
    factor: int = 4,
):
    """
    Get linear decimated data

    Parameters
    ----------
    fs : float, optional
        Sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 1024
    n_levels : int, optional
        The number of levels, by default 3
    factor : int, optional
        The decimation factor for each level, by default 4

    Returns
    -------
    DecimatedData
        The decimated data
    """
    metadata = decimated_metadata(
        fs, first_time, n_samples=n_samples, n_levels=n_levels, factor=factor
    )
    data = {}
    for ilevel in range(metadata.n_levels):
        level_samples = metadata.levels_metadata[ilevel].n_samples
        level_data = np.array([np.arange(level_samples), -1 * np.arange(level_samples)])
        data[ilevel] = level_data
    creator = {
        "name": "decimated_data_linear",
        "fs": fs,
        "first_time": first_time,
        "n_levels": n_levels,
    }
    record = get_record(creator, "Generated linear decimated data")
    metadata.history.add_record(record)
    return DecimatedData(metadata, data)


def decimated_data_periodic(
    frequencies: Dict[str, List[float]],
    fs: float = 0.25,
    first_time: str = "2021-01-01 00:00:00",
    n_samples: int = 1024,
    n_levels: int = 3,
    factor: int = 4,
):
    """
    Get periodic decimated data

    Parameters
    ----------
    frequencies : Dict[str, List[float]]
        Mapping from channel to list of frequencies to add
    fs : float, optional
        Sampling frequency, by default 10
    first_time : str, optional
        The time of the first sample, by default "2021-01-01 00:00:00"
    n_samples : int, optional
        The number of samples, by default 1024
    n_levels : int, optional
        The number of levels, by default 3
    factor : int, optional
        The decimation factor for each level, by default 4

    Returns
    -------
    DecimatedData
        The decimated data
    """
    metadata = decimated_metadata(
        fs, first_time, n_samples=n_samples, n_levels=n_levels, factor=factor
    )
    data = {}
    for ilevel in range(metadata.n_levels):
        level_samples = metadata.levels_metadata[ilevel].n_samples
        level_fs = metadata.levels_metadata[ilevel].fs
        times = np.arange(0, level_samples) * (1 / level_fs)
        level_data = []
        for chan in metadata.chans:
            chan_data = np.zeros(shape=(level_samples))
            for freq in frequencies[chan]:
                if freq > level_fs / 2:
                    continue
                chan_data += np.sin(times * 2 * np.pi * freq)
            level_data.append(chan_data)
        data[ilevel] = np.array(level_data)
    creator = {
        "name": "decimated_data_periodic",
        "fs": fs,
        "first_time": first_time,
        "n_levels": n_levels,
    }
    record = get_record(creator, "Generated periodic decimated data")
    metadata.history.add_record(record)
    return DecimatedData(metadata, data)


def spectra_metadata_multilevel(
    fs: float = 128,
    n_levels: int = 3,
    n_wins: Union[List[int], int] = 2,
    index_offset: Union[List[int], int] = 0,
    chans: Optional[List[str]] = None,
) -> SpectraMetadata:
    """
    Get spectra metadata with multiple levels and two channels

    Parameters
    ----------
    fs : float, optional
        The original sampling frequency, by default 128
    n_levels : int, optional
        The number of levels, by default 3
    n_wins: Union[List[int], int]
        The number of windows for each level
    index_offset : Union[List[int], int], optional
        The index offset vs. the reference time, by default 0
    chans : Optional[List[str]]
        The channels in the data, by default None. If None, the channels will be
        chan1 and chan2

    Returns
    -------
    SpectraMetadata
        SpectraMetadata with n_levels

    Raises
    ------
    ValueError
        If the number of user input channels does not equal two
    """
    if isinstance(n_wins, int):
        n_wins = (n_wins * np.ones(shape=(n_levels))).tolist()
    if isinstance(index_offset, int):
        index_offset = (index_offset * np.ones(shape=(n_levels))).tolist()

    levels_metadata = []
    levels_fs = []
    for ilevel, offset in zip(range(n_levels), index_offset):
        factor = np.power(2, ilevel)
        fs = fs / factor
        levels_metadata.append(
            SpectraLevelMetadata(
                fs=fs,
                n_wins=n_wins[ilevel],
                win_size=20,
                olap_size=5,
                index_offset=offset,
                n_freqs=2,
                freqs=[fs / 4, fs / 8],
            )
        )
        levels_fs.append(fs)
    metadata_dict = time_metadata_2chan().dict()
    if chans is not None:
        if len(chans) != 2:
            raise ValueError(f"More than two channels {chans}")
        metadata_dict["chans"] = chans
    metadata_dict["fs"] = levels_fs
    metadata_dict["n_levels"] = len(levels_metadata)
    metadata_dict["levels_metadata"] = levels_metadata
    metadata_dict["ref_time"] = metadata_dict["first_time"]
    return SpectraMetadata(**metadata_dict)


def spectra_data_basic() -> SpectraData:
    """
    Spectra data with a single decimation level

    Returns
    -------
    SpectraData
        Spectra data with a single level, a single channel and two windows
    """

    data = {}
    # fmt:off
    data[0] = np.array(
        [
            [[0 + 0j, 1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j, 5 + 5j, 6 + 6j, 7 + 7j, 8 + 8j, 9 + 9j]],
            [[-1 + 1j, 0 + 2j, 1 + 3j, 2 + 4j, 3 + 5j, 4 + 6j, 5 + 7j, 6 + 8j, 7 + 9j, 8 + 10j]],
        ]
    )
    # fmt:on
    freqs = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]
    level_metadata = SpectraLevelMetadata(
        fs=180,
        n_wins=2,
        win_size=20,
        olap_size=5,
        index_offset=0,
        n_freqs=10,
        freqs=freqs,
    )
    metadata_dict = time_metadata_1chan().dict()
    metadata_dict["fs"] = [180]
    metadata_dict["n_levels"] = 1
    metadata_dict["levels_metadata"] = [level_metadata]
    metadata_dict["ref_time"] = metadata_dict["first_time"]
    metadata = SpectraMetadata(**metadata_dict)
    creator = {
        "name": "spec_data_basic",
    }
    record = get_record(creator, "Generated spectra data with 1 channel and 1 level")
    metadata.history.add_record(record)
    return SpectraData(metadata, data)


def regression_input_metadata_mt(
    fs: float, freqs: List[float]
) -> RegressionInputMetadata:
    """
    Get example regression input metadata for single site mt

    Parameters
    ----------
    fs : float
        The sampling frequency
    freqs : List[float]
        The evaluation frequencies

    Returns
    -------
    RegressionInputMetadata
        Example regression input metadata with fs=128 and 5 evaluation
        frequencies
    """
    out_site = SiteCombinedMetadata(
        site_name="site1",
        fs=fs,
        measurements=["run1", "run2"],
        chans=["Ex", "Ey"],
        n_evals=len(freqs),
        eval_freqs=freqs,
        histories={"run1": History(), "run2": History()},
    )
    in_site = SiteCombinedMetadata(
        site_name="site1",
        fs=fs,
        measurements=["run1", "run2"],
        chans=["Hx", "Hy"],
        n_evals=len(freqs),
        eval_freqs=freqs,
        histories={"run1": History(), "run2": History()},
    )
    cross_site = SiteCombinedMetadata(**in_site.dict())
    creator = {
        "name": "regression_input_metadata_mt",
    }
    record = get_record(creator, "Generated testing regression input metadata for MT")
    return RegressionInputMetadata(
        contributors={
            "out_data": out_site,
            "in_data": in_site,
            "cross_data": cross_site,
        },
        history=History(records=[record]),
    )


def components_mt() -> Dict[str, Component]:
    """
    Get example components for the Impedance Tensor

    Returns
    -------
    Dict[str, Component]
        Dictionary of component values (ExHx, ExHy, EyHx, EyHy)
    """
    components = {
        "ExHx": Component(real=[1, 1, 2, 2, 3], imag=[5, 5, 4, 4, 3]),
        "ExHy": Component(real=[1, 2, 3, 4, 5], imag=[-5, -4, -3, -2, -1]),
        "EyHx": Component(real=[-1, -2, -3, -4, -5], imag=[5, 4, 3, 2, 1]),
        "EyHy": Component(real=[-1, -1, -2, -2, -3], imag=[-5, -5, -4, -4, -3]),
    }
    return components


def solution_mt() -> Solution:
    """
    Get an example impedance tensor solution

    Returns
    -------
    Solution
        The solution
    """
    tf = ImpedanceTensor()
    fs = 128
    freqs = [10.0, 20.0, 30.0, 40.0, 50.0]
    components = components_mt()
    metadata = regression_input_metadata_mt(fs, freqs)
    return Solution(
        tf=tf,
        freqs=freqs,
        components=components,
        history=History(),
        contributors=metadata.contributors,
    )



"""
Common resistics functions and classes used throughout the package
"""
from loguru import logger
from typing import List, Tuple, Union, Dict
from typing import Any, Collection, Optional, Type
from pathlib import Path
from pydantic import BaseModel, Field, validator
from datetime import datetime
import numpy as np

from resistics.sampling import RSDateTime, datetime_to_string

ELECTRIC_CHANS = ["Ex", "Ey", "E1", "E2", "E3", "E4"]
MAGNETIC_CHANS = ["Hx", "Hy", "Hz", "Bx", "By", "Bz"]


def get_version() -> str:
    """Get the version of resistics"""
    import resistics

    return resistics.__version__


def is_file(file_path: Path) -> bool:
    """
    Check if a path exists and points to a file

    Parameters
    ----------
    file_path : Path
        The path to check

    Returns
    -------
    bool
        True if it exists and is a file, False otherwise
    """
    if not file_path.exists():
        logger.warning(f"File path {file_path} does not exist")
        return False
    if not file_path.is_file():
        logger.warning(f"File path {file_path} is not a file")
        return False
    return True


def assert_file(file_path: Path) -> None:
    """
    Require that a file exists

    Parameters
    ----------
    file_path : Path
        The path to check

    Raises
    ------
    FileNotFoundError
        If the path does not exist
    NotFileError
        If the path is not a file
    """
    from resistics.errors import NotFileError

    if not file_path.exists():
        raise FileNotFoundError(f"Path {file_path} not found")
    if not file_path.is_file():
        raise NotFileError(file_path)


def is_dir(dir_path: Path) -> bool:
    """
    Check if a path exists and points to a directory

    Parameters
    ----------
    dir_path : Path
        The path to check

    Returns
    -------
    bool
        True if it exists and is a directory, False otherwise
    """
    if not dir_path.exists():
        logger.warning(f"Directory path {dir_path} does not exist")
        return False
    if not dir_path.is_dir():
        logger.warning(f"Directory path {dir_path} is not a directory")
        return False
    return True


def assert_dir(dir_path: Path) -> None:
    """
    Require that a path is a directory

    Parameters
    ----------
    dir_path : Path
        Path to check

    Raises
    ------
    FileNotFoundError
        If the path does not exist
    NotDirectoryError
        If the path is not a directory
    """
    from resistics.errors import NotDirectoryError

    if not dir_path.exists():
        raise FileNotFoundError(f"Path {dir_path} does not exist")
    if not dir_path.is_dir():
        raise NotDirectoryError(dir_path)


def dir_contents(dir_path: Path) -> Tuple[List[Path], List[Path]]:
    """
    Get contents of directory

    Includes both files and directories

    Parameters
    ----------
    dir_path : Path
        Parent directory path

    Returns
    -------
    dirs : list
        List of directories
    files : list
        List of files excluding hidden files

    Raises
    ------
    PathNotFoundError
        Path does not exist
    NotDirectoryError
        Path is not a directory
    """
    from resistics.errors import PathNotFoundError, NotDirectoryError

    if not dir_path.exists():
        raise PathNotFoundError(dir_path)
    if not dir_path.is_dir():
        raise NotDirectoryError(dir_path)

    dirs = []
    files = []
    for obj in dir_path.iterdir():
        if obj.is_file():
            files.append(obj)
        elif obj.is_dir():
            dirs.append(obj)
    return dirs, files


def dir_files(dir_path: Path) -> List[Path]:
    """
    Get files in directory

    Excludes hidden files

    Parameters
    ----------
    dir_path : Path
        Parent directory path

    Returns
    -------
    files : list
        List of files excluding hidden files
    """
    _, files = dir_contents(dir_path)
    return files


def dir_subdirs(dir_path: Path) -> List[Path]:
    """
    Get subdirectories in directory

    Excludes hidden files

    Parameters
    ----------
    dir_path : Path
        Parent directory path

    Returns
    -------
    dirs : list
        List of subdirectories
    """
    dirs, _ = dir_contents(dir_path)
    return dirs


def is_electric(chan: str) -> bool:
    """
    Check if a channel is electric

    Parameters
    ----------
    chan : str
        Channel name

    Returns
    -------
    bool
        True if channel is electric

    Examples
    --------
    >>> from resistics.common import is_electric
    >>> is_electric("Ex")
    True
    >>> is_electric("Hx")
    False
    """
    if chan in ELECTRIC_CHANS:
        return True
    return False


def is_magnetic(chan: str) -> bool:
    """
    Check if channel is magnetic

    Parameters
    ----------
    chan : str
        Channel name

    Returns
    -------
    bool
        True if channel is magnetic

    Examples
    --------
    >>> from resistics.common import is_magnetic
    >>> is_magnetic("Ex")
    False
    >>> is_magnetic("Hx")
    True
    """
    if chan in MAGNETIC_CHANS:
        return True
    return False


def get_chan_type(chan: str) -> str:
    """
    Get the channel type from the channel name

    Parameters
    ----------
    chan : str
        The name of the channel

    Returns
    -------
    str
        The channel type

    Raises
    ------
    ValueError
        If the channel is not known to resistics

    Examples
    --------
    >>> from resistics.common import get_chan_type
    >>> get_chan_type("Ex")
    'electric'
    >>> get_chan_type("Hz")
    'magnetic'
    >>> get_chan_type("abc")
    Traceback (most recent call last):
    ...
    ValueError: Channel abc not recognised as either electric or magnetic
    """
    if is_electric(chan):
        return "electric"
    if is_magnetic(chan):
        return "magnetic"
    raise ValueError(f"Channel {chan} not recognised as either electric or magnetic")


def check_chan(chan: str, chans: Collection[str]) -> bool:
    """
    Check a channel exists and raise a KeyError if not

    Parameters
    ----------
    chan : str
        The channel to check
    chans : Collection[str]
        A collection of channels to check against

    Returns
    -------
    bool
        True if all checks passed

    Raises
    ------
    ChannelNotFoundError
        If the channel is not found in the channel list
    """
    from resistics.errors import ChannelNotFoundError

    if chan not in chans:
        logger.error(f"Channel {chan} not in channel list {chans}")
        raise ChannelNotFoundError(chan, chans)
    return True


def fs_to_string(fs: float) -> str:
    """
    Convert sampling frequency into a string for filenames

    Parameters
    ----------
    fs : float
        The sampling frequency

    Returns
    -------
    str
        Sample frequency converted to string for the purposes of a filename

    Examples
    --------
    >>> from resistics.common import fs_to_string
    >>> fs_to_string(512.0)
    '512_000000'
    """
    return (f"{fs:.6f}").replace(".", "_")


def array_to_string(
    data: np.ndarray, sep: str = ", ", precision: int = 8, scientific: bool = False
) -> str:
    """
    Convert an array to a string for logging or printing

    Parameters
    ----------
    data : np.ndarray
        The array
    sep : str, optional
        The separator to use, by default ", "
    precision : int, optional
        Number of decimal places, by default 8. Ignored for integers.
    scientific : bool, optional
        Flag for formatting floats as scientific, by default False

    Returns
    -------
    str
        String representation of array

    Examples
    --------
    >>> import numpy as np
    >>> from resistics.common import array_to_string
    >>> data = np.array([1,2,3,4,5])
    >>> array_to_string(data)
    '1, 2, 3, 4, 5'
    >>> data = np.array([1,2,3,4,5], dtype=np.float32)
    >>> array_to_string(data)
    '1.00000000, 2.00000000, 3.00000000, 4.00000000, 5.00000000'
    >>> array_to_string(data, precision=3, scientific=True)
    '1.000e+00, 2.000e+00, 3.000e+00, 4.000e+00, 5.000e+00'
    """
    style: str = "e" if scientific else "f"
    output_str = np.array2string(
        data,
        separator=sep,
        formatter={"float_kind": lambda x: f"{x:.{precision}{style}}"},
    )
    return output_str.lstrip("[").rstrip("]")


class ResisticsModel(BaseModel):
    """Base resistics model"""

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        """Class info as string"""
        import json
        import yaml

        json_dict = json.loads(self.json())
        return yaml.dump(json_dict, indent=4, sort_keys=False)

    def summary(self) -> None:
        """Print a summary of the class"""
        import json
        from prettyprinter import cpprint

        cpprint(json.loads(self.json()))

    class Config:
        """pydantic configuration information"""

        json_encoders = {RSDateTime: datetime_to_string}


class ResisticsFile(ResisticsModel):
    """Required information for writing out a resistics file"""

    created_on_local: datetime = Field(default_factory=datetime.now)
    created_on_utc: datetime = Field(default_factory=datetime.utcnow)
    version: Optional[str] = Field(default_factory=get_version)


class Metadata(ResisticsModel):
    """Parent class for metadata"""

    @validator("n_chans", check_fields=False, always=True)
    def validate_n_chans(cls, value: Union[None, int], values: Dict[str, Any]) -> int:
        """Initialise number of channels"""
        if value is None:
            return len(values["chans"])
        return value


class WriteableMetadata(Metadata):
    """Base class for writeable metadata"""

    file_info: Optional[ResisticsFile] = None
    """Information about a file, relevant if writing out or reading back in"""

    def write(self, json_path: Path):
        """
        Write out JSON metadata file

        Parameters
        ----------
        json_path : Path
            Path to write JSON file
        """
        self.file_info = ResisticsFile()
        with json_path.open("w") as f:
            f.write(self.json())


class Record(ResisticsModel):
    """
    Class to hold a record

    A record holds information about a process that was run. It is intended to
    track processes applied to data, allowing a process history to be saved
    along with any datasets.

    Examples
    --------
    A simple example of creating a process record

    >>> from resistics.common import Record
    >>> messages = ["message 1", "message 2"]
    >>> record = Record(
    ...     creator={"name": "example", "parameter1": 15},
    ...     messages=messages,
    ...     record_type="example"
    ... )
    >>> record.summary()
    {
        'time_local': '...',
        'time_utc': '...',
        'creator': {'name': 'example', 'parameter1': 15},
        'messages': ['message 1', 'message 2'],
        'record_type': 'example'
    }
    """

    time_local: datetime = Field(default_factory=datetime.now)
    """The local time when the process ran"""
    time_utc: datetime = Field(default_factory=datetime.utcnow)
    """The UTC time when the process ran"""
    creator: Dict[str, Any]
    """The creator and its parameters as a dictionary"""
    messages: List[str]
    """Any messages in the record"""
    record_type: str
    """The record type"""


class History(ResisticsModel):
    """
    Class for storing processing history

    Parameters
    ----------
    records : List[Record], optional
        List of records, by default []

    Examples
    --------
    >>> from resistics.testing import record_example1, record_example2
    >>> from resistics.common import History
    >>> record1 = record_example1()
    >>> record2 = record_example2()
    >>> history = History(records=[record1, record2])
    >>> history.summary()
    {
        'records': [
            {
                'time_local': '...',
                'time_utc': '...',
                'creator': {
                    'name': 'example1',
                    'a': 5,
                    'b': -7.0
                },
                'messages': ['Message 1', 'Message 2'],
                'record_type': 'process'
            },
            {
                'time_local': '...',
                'time_utc': '...',
                'creator': {
                    'name': 'example2',
                    'a': 'parzen',
                    'b': -21
                },
                'messages': ['Message 5', 'Message 6'],
                'record_type': 'process'
            }
        ]
    }
    """

    records: List[Record] = []

    def add_record(self, record: Record):
        """
        Add a process record to the list

        Parameters
        ----------
        record : Record
            The record to add
        """
        self.records.append(record)


def get_record(
    creator: Dict[str, Any],
    messages: Union[str, List[str]],
    record_type: str = "process",
    time_utc: Optional[datetime] = None,
    time_local: Optional[datetime] = None,
) -> Record:
    """
    Get a process record

    Parameters
    ----------
    creator : Dict[str, Any]
        The creator and its parameters as a dictionary
    messages : Union[str, List[str]]
        The messages as either a single str or a list of strings
    record_type : str, optional
        The type of record, by default "process"
    time_utc : Optional[datetime], optional
        UTC time to attach to the record, by default None. If None, will default
        to UTC now
    time_local : Optional[datetime], optional
        Local time to attach to the record, by default None. If None, will
        defult to local now

    Returns
    -------
    Record
        The process record

    Examples
    --------
    >>> from resistics.common import get_record
    >>> record = get_record(
    ...     creator={"name": "example", "a": 5, "b": -7.0},
    ...     messages="a message"
    ... )
    >>> record.creator
    {'name': 'example', 'a': 5, 'b': -7.0}
    >>> record.messages
    ['a message']
    >>> record.record_type
    'process'
    >>> record.time_utc
    datetime.datetime(...)
    >>> record.time_local
    datetime.datetime(...)
    """
    if isinstance(messages, str):
        messages = [messages]
    if time_utc is None:
        time_utc = datetime.utcnow()
    if time_local is None:
        time_local = datetime.now()
    return Record(
        time_utc=time_utc,
        time_local=time_local,
        creator=creator,
        messages=messages,
        record_type=record_type,
    )


def get_history(record: Record, history: Optional[History] = None) -> History:
    """
    Get a new History instance or add a record to a copy of an existing one

    This method always makes a deepcopy of an input history to avoid any
    unplanned modifications to the inputs.

    Parameters
    ----------
    record : Record
        The record
    history : Optional[History], optional
        A history to add to, by default None

    Returns
    -------
    History
        History with the record added

    Examples
    --------
    Get a new History with a single Record

    >>> from resistics.common import get_history
    >>> from resistics.testing import record_example1, record_example2
    >>> record1 = record_example1()
    >>> history = get_history(record1)
    >>> history.summary()
    {
        'records': [
            {
                'time_local': '...',
                'time_utc': '...',
                'creator': {
                    'name': 'example1',
                    'a': 5,
                    'b': -7.0
                },
                'messages': ['Message 1', 'Message 2'],
                'record_type': 'process'
            }
        ]
    }

    Alternatively, add to an existing History. This will make a copy of the
    original history. If a copy is not needed, the add_record method of history
    can be used.

    >>> record2 = record_example2()
    >>> history = get_history(record2, history)
    >>> history.summary()
    {
        'records': [
            {
                'time_local': '...',
                'time_utc': '...',
                'creator': {
                    'name': 'example1',
                    'a': 5,
                    'b': -7.0
                },
                'messages': ['Message 1', 'Message 2'],
                'record_type': 'process'
            },
            {
                'time_local': '...',
                'time_utc': '...',
                'creator': {
                    'name': 'example2',
                    'a': 'parzen',
                    'b': -21
                },
                'messages': ['Message 5', 'Message 6'],
                'record_type': 'process'
            }
        ]
    }
    """
    if history is None:
        return History(records=[record])
    history = History(**history.dict())
    history.add_record(record)
    return history


class ResisticsProcess(ResisticsModel):
    """
    Base class for resistics processes

    Resistics processes perform operations on data (including read and write
    operations). Each time a ResisticsProcess child class is run, it should add
    a process record to the dataset
    """

    _types: Dict[str, type] = {}
    name: Optional[str]

    def __init_subclass__(cls) -> None:
        """
        Used to automatically register child processors in `_types`

        When a resistics process is imported, it is added to the base
        ResisticsProcess _types variable. Later, this dictionary of class types
        can be used to initialise processes from a dictonary.

        The intention of this method is to support initialising processes from
        JSON files.
        """
        cls._types[cls.__name__] = cls

    @classmethod
    def __get_validators__(cls):
        """Get the validators that will be used by pydantic"""
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Union["ResisticsProcess", Dict[str, Any]]
    ) -> "ResisticsProcess":
        """
        Validate a ResisticsProcess in another pydantic class

        Parameters
        ----------
        value : Union[ResisticsProcess, Dict[str, Any]]
            A ResisticsProcess child class or a dictionary

        Returns
        -------
        ResisticsProcess
            A ResisticsProcess child class

        Raises
        ------
        ValueError
            If the value is neither a ResisticsProcess or a dictionary
        KeyError
            If name is not in the dictionary
        ValueError
            If initialising from dictionary fails

        Examples
        --------
        The following example will show how a generic ResisticsProcess child
        class can be instantiated from ResisticsProcess using a dictionary,
        which might be read in from a JSON configuration file.

        >>> from resistics.common import ResisticsProcess
        >>> from resistics.decimate import DecimationSetup
        >>> process = {"name": 'DecimationSetup', "n_levels": 8, "per_level": 5, "min_samples": 256, "div_factor": 2, "eval_freqs": None}
        >>> ResisticsProcess(**process)
        ResisticsProcess(name='DecimationSetup')

        This is not what was expected. To get the right result, the class
        validate method needs to be used. This is done automatically by
        pydantic.

        >>> ResisticsProcess.validate(process)
        DecimationSetup(name='DecimationSetup', n_levels=8, per_level=5, min_samples=256, div_factor=2, eval_freqs=None)

        That's better. Note that errors will be raised if the dictionary is not
        formatted as expected.

        >>> process = {"n_levels": 8, "per_level": 5, "min_samples": 256, "div_factor": 2, "eval_freqs": None}
        >>> ResisticsProcess.validate(process)
        Traceback (most recent call last):
        ...
        KeyError: 'No name provided for initialisation of process'

        This functionality is most useful in the resistics configurations which
        can be saved as JSON files. The default configuration uses the default
        parameterisation of DecimationSetup.

        >>> from resistics.letsgo import Configuration
        >>> config = Configuration(name="example1")
        >>> config.dec_setup
        DecimationSetup(name='DecimationSetup', n_levels=8, per_level=5, min_samples=256, div_factor=2, eval_freqs=None)

        Now create another configuration with a different setup by passing a
        dictionary. In practise, this dictionary will most likely be read in
        from a configuration file.

        >>> setup = DecimationSetup(n_levels=4, per_level=3)
        >>> test_dict = setup.dict()
        >>> test_dict
        {'name': 'DecimationSetup', 'n_levels': 4, 'per_level': 3, 'min_samples': 256, 'div_factor': 2, 'eval_freqs': None}
        >>> config2 = Configuration(name="example2", dec_setup=test_dict)
        >>> config2.dec_setup
        DecimationSetup(name='DecimationSetup', n_levels=4, per_level=3, min_samples=256, div_factor=2, eval_freqs=None)

        This method allows the saving of a configuration with custom processors
        in a JSON file which can be loaded and used again.
        """
        if isinstance(value, ResisticsProcess):
            return value
        if not isinstance(value, dict):
            raise ValueError(
                "ResisticsProcess unable to initialise from type {type(value)}"
            )
        if "name" not in value:
            raise KeyError("No name provided for initialisation of process")
        name = value.pop("name")
        try:
            return cls._types[name](**value)
        except Exception:
            raise ValueError(f"Unable to initialise {name} from dictionary")

    @validator("name", always=True)
    def validate_name(cls, value: Union[str, None]) -> str:
        """Inialise the name attribute of the resistics process"""
        if value is None:
            return cls.__name__
        return value

    def parameters(self) -> Dict[str, Any]:
        """
        Return any process parameters incuding the process name

        These parameters are expected to be primatives and should be sufficient
        to reinitialise the process and re-run the data. The base class assumes
        all class variables meet this description.

        Returns
        -------
        Dict[str, Any]
            Dictionary of parameters
        """
        import json

        return json.loads(self.json())

    def _get_record(self, messages: Union[str, List[str]]) -> Record:
        """
        Get the record for the processor

        Parameters
        ----------
        messages : Union[str, List[str]]
            The messages to add for the processor

        Returns
        -------
        Record
            A record
        """
        return get_record(self.parameters(), messages)


class ResisticsBase(object):
    """
    Resistics base class

    Parent class to ensure consistency of common methods
    """

    def __repr__(self) -> str:
        """Return a string of class information"""
        return self.to_string()

    def __str__(self) -> str:
        """Return a string of class information"""
        return self.to_string()

    def type_to_string(self) -> str:
        """Get the class type as a string"""
        return str(self.__class__)

    def to_string(self) -> str:
        """Class details as a string"""
        return self.type_to_string()

    def summary(self, symbol: str = "-") -> None:
        """Print a summary of class details"""
        name = str(self.__class__)
        length = len(name) + 10
        print("##" + 3 * symbol + "Begin Summary" + ((length - 18) * symbol))
        print(self.to_string())
        print("##" + 3 * symbol + "End summary" + (length - 16) * symbol)


class ResisticsData(ResisticsBase):
    """Base class for a resistics data object"""

    pass


class ResisticsWriter(ResisticsProcess):
    """
    Parent process for data writers

    Parameters
    ----------
    overwrite : bool, optional
        Boolean flag for overwriting the existing data, by default False
    """

    overwrite: bool = True

    def run(self, dir_path: Path, data: ResisticsData) -> None:
        """Write out a ResisticsData child object to a directory"""
        raise NotImplementedError("To be implemented in child writers")

    def _check_dir(self, dir_path: Path) -> bool:
        """Check the output directory"""
        if dir_path.exists() and not self.overwrite:
            logger.error(f"Write path {dir_path} exists and overwrite is False")
            return False
        if dir_path.exists():
            logger.warning(f"Overwriting existing directory {dir_path}")
        if not dir_path.exists():
            logger.info(f"Directory {dir_path} not found. Creating including parents.")
            dir_path.mkdir(parents=True)
        return True

    def _get_record(self, dir_path: Path, data_type: Type):
        """Get a process record for the writer"""
        return super()._get_record([f"Writing out {data_type.__name__} to {dir_path}"])



"""
Classes and methods to enable a resistics project

A project is an essential element of a resistics environment together with a
configuration.

In particular, this module includes the core Project, Site and Measurement
clasess and some supporting functions.
"""
from loguru import logger
from typing import Iterator, Optional, List, Dict
from pathlib import Path
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px

from resistics.common import ResisticsModel, WriteableMetadata
from resistics.sampling import HighResDateTime
from resistics.time import TimeMetadata
from resistics.time import TimeReader


PROJ_FILE = "resistics.json"
PROJ_DIRS = {
    "time": "time",
    "calibration": "calibrate",
    "spectra": "spectra",
    "evals": "evals",
    "features": "features",
    "masks": "masks",
    "results": "results",
    "images": "images",
}


def get_calibration_path(proj_dir: Path) -> Path:
    """Get the path to the calibration data"""
    return proj_dir / PROJ_DIRS["calibration"]


def get_meas_time_path(proj_dir: Path, site_name: str, meas_name: str) -> Path:
    """Get path to measurement time data"""
    return proj_dir / PROJ_DIRS["time"] / site_name / meas_name


def get_meas_spectra_path(
    proj_dir: Path, site_name: str, meas_name: str, config_name: str
) -> Path:
    """Get path to measurement spectra data"""
    return proj_dir / PROJ_DIRS["spectra"] / site_name / config_name / meas_name


def get_meas_evals_path(
    proj_dir: Path, site_name: str, meas_name: str, config_name: str
) -> Path:
    """Get path to measurement evaluation frequency spectra data"""
    return proj_dir / PROJ_DIRS["evals"] / site_name / config_name / meas_name


def get_meas_features_path(
    proj_dir: Path, site_name: str, meas_name: str, config_name: str
) -> Path:
    """Get path to measurement features data"""
    return proj_dir / PROJ_DIRS["features"] / site_name / config_name / meas_name


def get_mask_path(proj_dir: Path, site_name: str, config_name: str) -> Path:
    """Get path to mask data"""
    return proj_dir / PROJ_DIRS["masks"] / site_name / config_name


def get_mask_name(fs: float, mask_name: str) -> str:
    """Get the name of a mask file"""
    from resistics.common import fs_to_string

    return f"{fs_to_string(fs)}_{mask_name}.dat"


def get_results_path(proj_dir: Path, site_name: str, config_name: str) -> Path:
    """Get path to solutions"""
    return proj_dir / PROJ_DIRS["results"] / site_name / config_name


def get_solution_name(
    fs: float, tf_name: str, tf_var: str, postfix: Optional[str] = None
) -> str:
    """Get the name of a solution file"""
    from resistics.common import fs_to_string

    solution_name = f"{fs_to_string(fs)}_{tf_name.lower()}"
    if tf_var != "":
        tf_var = tf_var.replace(" ", "_")
        solution_name = solution_name + f"_{tf_var}"
    if postfix is None:
        return solution_name + ".json"
    return solution_name + "_" + postfix + ".json"


class Measurement(ResisticsModel):
    """
    Class for interfacing with a measurement

    The class holds the original time series metadata and can provide
    information about other types of data
    """

    site_name: str
    dir_path: Path
    metadata: TimeMetadata
    reader: TimeReader

    @property
    def name(self) -> str:
        """Get the name of the measurement"""
        return self.dir_path.name


class Site(ResisticsModel):
    """
    Class for describing Sites

    .. note::

        This should essentially describe a single instrument setup. If the same
        site is re-occupied later with a different instrument setup, it is
        suggested to split this into a different site.
    """

    dir_path: Path
    measurements: Dict[str, Measurement]
    begin_time: HighResDateTime
    end_time: HighResDateTime

    def __iter__(self) -> Iterator:
        """Iterator over measurements"""
        return self.measurements.values().__iter__()

    def __getitem__(self, meas_name: str) -> Measurement:
        """Get a measurement"""
        return self.get_measurement(meas_name)

    @property
    def name(self) -> str:
        """The Site name"""
        return self.dir_path.name

    @property
    def n_meas(self) -> int:
        """Get the number of measurements"""
        return len(self.measurements)

    def fs(self) -> List[float]:
        """Get the sampling frequencies in the Site"""
        fs = [x.metadata.fs for x in self.measurements.values()]
        return sorted(list(set(fs)))

    def get_measurement(self, meas_name: str) -> Measurement:
        """Get a measurement"""
        from resistics.errors import MeasurementNotFoundError

        if meas_name not in self.measurements:
            raise MeasurementNotFoundError(self.name, meas_name)
        return self.measurements[meas_name]

    def get_measurements(self, fs: Optional[float] = None) -> Dict[str, Measurement]:
        """Get dictionary of measurements with optional filter by sampling frequency"""
        if fs is None:
            return self.measurements
        return {
            name: meas
            for name, meas in self.measurements.items()
            if meas.metadata.fs == fs
        }

    def plot(self) -> go.Figure:
        """Plot the site timeline"""
        df = self.to_dataframe()
        if len(df.index) == 0:
            logger.error("No measurements found to plot")
            return
        fig = px.timeline(
            df,
            x_start="first_time",
            x_end="last_time",
            y="name",
            color="fs",
            title=self.name,
        )
        return fig

    def to_dataframe(self) -> pd.DataFrame:
        """
        Get measurements list in a pandas DataFrame

        .. note::

            Measurement first and last times are converted to pandas Timestamps
            as these are more universally useful in a pandas DataFrame. However,
            this may result in a loss of precision, especially at high sampling
            frequencies.

        Returns
        -------
        pd.DataFrame
            Site measurement DataFrame
        """
        data = [
            [
                x.name,
                x.metadata.fs,
                x.metadata.first_time.isoformat(),
                x.metadata.last_time.isoformat(),
            ]
            for x in self.measurements.values()
        ]
        df = pd.DataFrame(data=data, columns=["name", "fs", "first_time", "last_time"])
        df["first_time"] = pd.to_datetime(df["first_time"])
        df["last_time"] = pd.to_datetime(df["last_time"])
        df["site"] = self.name
        return df


class ProjectMetadata(WriteableMetadata):
    """Project metadata"""

    ref_time: HighResDateTime
    location: str = ""
    country: str = ""
    year: int = -999
    description: str = ""
    contributors: List[str] = []


class Project(ResisticsModel):
    """
    Class to describe a resistics project

    The resistics Project Class connects all resistics data. It is an essential
    part of processing data with resistics.

    Resistics projects are in directory with several sub-directories. Project
    metadata is saved in the resistics.json file at the top level directory.
    """

    dir_path: Path
    begin_time: HighResDateTime
    end_time: HighResDateTime
    metadata: ProjectMetadata
    sites: Dict[str, Site] = {}

    def __iter__(self) -> Iterator:
        """Iterator over sites"""
        return self.sites.values().__iter__()

    def __getitem__(self, site_name: str) -> Site:
        """Get a Site instance given the name of the Site"""
        return self.get_site(site_name)

    @property
    def n_sites(self) -> int:
        """The number of sites"""
        return len(self.sites)

    def fs(self) -> List[float]:
        """Get sampling frequencies in the Project"""
        fs = set()
        for site in self.sites.values():
            fs = fs.union(set(site.fs()))
        return sorted(list(fs))

    def get_site(self, site_name: str) -> Site:
        """Get a Site object given the Site name"""
        from resistics.errors import SiteNotFoundError

        if site_name not in self.sites:
            raise SiteNotFoundError(site_name)
        return self.sites[site_name]

    def get_sites(self, fs: Optional[float] = None) -> Dict[str, Site]:
        """
        Get sites

        Parameters
        ----------
        fs : Optional[float], optional
            Filter by sites which have at least a single recording at a
            specified sampling frequency, by default None

        Returns
        -------
        Dict[str, Site]
            Dictionary of site name to Site
        """
        if fs is None:
            return self.sites
        return {name: site for name, site in self.sites.items() if fs in site.fs()}

    def get_concurrent(self, site_name: str) -> List[Site]:
        """
        Find sites that recorded conscurrently to a specified site

        Parameters
        ----------
        site_name : str
            Search for sites recording concurrently to this site

        Returns
        -------
        List[Site]
            List of Site instances which were recording concurrently
        """
        site_start = self.sites[site_name].begin_time
        site_end = self.sites[site_name].end_time
        concurrent = []
        for site in self.sites.values():
            if site.name == site_name:
                continue
            if site.end_time < site_start:
                continue
            if site.begin_time > site_end:
                continue
            concurrent.append(site)
        return concurrent

    def plot(self) -> go.Figure:
        """Plot a timeline of the project"""
        df = self.to_dataframe()
        df["Site"] = df["site"]
        df["Sampling frequency, Hz"] = df["fs"].values.astype(str)
        if len(df.index) == 0:
            logger.error("No measurements found to plot")
            return
        fig = px.timeline(
            df,
            x_start="first_time",
            x_end="last_time",
            y="Site",
            color="Sampling frequency, Hz",
            title=str(self.dir_path),
        )
        return fig

    def to_dataframe(self) -> pd.DataFrame:
        """Detail Project recordings in a DataFrame"""
        df = pd.DataFrame(columns=["name", "first_time", "last_time", "fs", "site"])
        for site in self.sites.values():
            df = df.append(site.to_dataframe())
        return df



"""
Module for time data decimation including classes and for the following

- Definition of DecimationParameters
- Performing decimation on time data
"""
from loguru import logger
from typing import Any, Optional, Tuple, Union, Dict, List
from pathlib import Path
from pydantic import validator, PositiveInt, conint
import numpy as np
import pandas as pd
import plotly.graph_objects as go

from resistics.common import ResisticsProcess, ResisticsModel, ResisticsData
from resistics.common import ResisticsWriter, History, Metadata, WriteableMetadata
from resistics.sampling import HighResDateTime
from resistics.time import ChanMetadata, TimeData


def get_eval_freqs_min(fs: float, f_min: float) -> np.ndarray:
    """
    Calculate evaluation frequencies with mimum allowable frequency

    Highest frequency is nyquist / 4

    Parameters
    ----------
    fs : float
        Sampling frequency
    f_min : float
        Minimum allowable frequency

    Returns
    -------
    np.ndarray
        Array of evaluation frequencies

    Raises
    ------
    ValueError
        If f_min <= 0

    Examples
    --------

    >>> from resistics.decimate import get_eval_freqs_min
    >>> fs = 256
    >>> get_eval_freqs_min(fs, 30)
    array([64.      , 45.254834, 32.      ])
    >>> get_eval_freqs_min(fs, 128)
    Traceback (most recent call last):
    ...
    ValueError: Minimum frequency 128 must be > 64.0
    """
    f0 = fs / 4

    if f_min <= 0:
        raise ValueError(f"Minimimum frequency {f_min} not > 0")
    if f_min > f0:
        raise ValueError(f"Minimum frequency {f_min} must be > {f0}")

    ii = 1
    eval_freqs = []
    while True:
        freq = f0 / np.power(2, (ii - 1.0) / 2.0)
        if freq < f_min:
            break
        eval_freqs.append(freq)
        ii += 1
    return np.array(eval_freqs)


def get_eval_freqs_size(fs: float, n_freqs: int) -> np.ndarray:
    """
    Calculate evaluation frequencies with maximum size

    Highest frequency is nyquist/4

    Parameters
    ----------
    fs : float
        Sampling frequency
    n_freqs : int
        Number of evaluation frequencies

    Returns
    -------
    np.ndarray
        Array of evaluation frequencies

    Examples
    --------
    >>> from resistics.decimate import get_eval_freqs_size
    >>> fs = 256
    >>> n_freqs = 3
    >>> get_eval_freqs_size(fs, n_freqs)
    array([64.      , 45.254834, 32.      ])
    """
    f0 = fs / 4
    return f0 / np.power(2, (np.arange(1, n_freqs + 1) - 1) / 2)


def get_eval_freqs(
    fs: float, f_min: Optional[float] = None, n_freqs: Optional[int] = None
) -> np.ndarray:
    """
    Get evaluation frequencies either based on size or a minimum frequency

    Parameters
    ----------
    fs : float
        Sampling frequency Hz
    f_min : Optional[float], optional
        Minimum cutoff for evaluation frequencies, by default None
    n_freqs : Optional[int], optional
        Number of evaluation frequencies, by default None

    Returns
    -------
    np.ndarray
        Evaluation frequencies array

    Raises
    ------
    ValueError
        ValueError if both f_min and n_freqs are None

    Examples
    --------
    >>> from resistics.decimate import get_eval_freqs
    >>> get_eval_freqs(256, f_min=30)
    array([64.      , 45.254834, 32.      ])
    >>> get_eval_freqs(256, n_freqs=3)
    array([64.      , 45.254834, 32.      ])
    """
    if f_min is None and n_freqs is None:
        raise ValueError("One of f_min and n_freqs must be passed")
    elif f_min is not None:
        return get_eval_freqs_min(fs, f_min)
    else:
        return get_eval_freqs_size(fs, n_freqs)


class DecimationParameters(ResisticsModel):
    """
    Decimation parameters

    Parameters
    ----------
    fs : float
        Sampling frequency Hz
    n_levels : int
        Number of levels
    per_level : int
        Evaluation frequencies per level
    min_samples : int
        Number of samples to under which to quit decimating
    eval_df : pd.DataFrame
        The DataFrame with the decimation information

    Examples
    --------
    >>> from resistics.decimate import DecimationSetup
    >>> dec_setup = DecimationSetup(n_levels=3, per_level=2)
    >>> dec_params = dec_setup.run(128)
    >>> type(dec_params)
    <class 'resistics.decimate.DecimationParameters'>
    >>> print(dec_params.to_dataframe())
                         0          1     fs  factors  increments
    decimation level
    0                 32.0  22.627417  128.0        1           1
    1                 16.0  11.313708   64.0        2           2
    2                  8.0   5.656854   32.0        4           2
    >>> dec_params[2]
    [8.0, 5.65685424949238]
    >>> dec_params[2,1]
    5.65685424949238
    >>> dec_params.get_total_factor(2)
    4
    >>> dec_params.get_incremental_factor(2)
    2
    """

    fs: float
    n_levels: int
    per_level: int
    min_samples: PositiveInt
    eval_freqs: List[float]
    dec_factors: List[int]
    dec_increments: Optional[List[int]] = None
    dec_fs: Optional[List[float]] = None

    @validator("dec_increments", always=True)
    def set_dec_increments(cls, value, values):
        """Initialise decimation increments if not provided"""
        if value is None:
            divisor = np.ones(shape=(values["n_levels"]), dtype=int)
            divisor[1:] = values["dec_factors"][:-1]
            return np.divide(values["dec_factors"], divisor).astype(int).tolist()
        return value

    @validator("dec_fs", always=True)
    def set_dec_fs(cls, value, values):
        """Initialise decimation sampling frequencies if not provided"""
        if value is None:
            factors = np.array(values["dec_factors"]).astype(float)
            return (values["fs"] * np.reciprocal(factors)).tolist()
        return value

    def __getitem__(self, args: Union[int, Tuple[int, int]]):
        """Get the evaluation frequency for level and evaluation frequency index"""
        if isinstance(args, int):
            return self.get_eval_freqs(args)
        level, idx = args
        return self.get_eval_freq(level, idx)

    def check_level(self, level: int):
        """Check level"""
        if level < 0 or level >= self.n_levels:
            raise ValueError(f"Level {level} not 0 <= level < {self.n_levels}")

    def check_eval_idx(self, idx: int):
        """Check evaluation frequency index"""
        if idx < 0 or idx >= self.per_level:
            raise ValueError(f"Index {idx} not 0 <= index < {self.per_level}")

    def get_eval_freqs(self, level: int) -> List[float]:
        """
        Get the evaluation frequencies for a level

        Parameters
        ----------
        level : int
            The decimation level

        Returns
        -------
        List[float]
            List of evaluation frequencies
        """
        self.check_level(level)
        index_from = self.per_level * level
        return self.eval_freqs[index_from : index_from + self.per_level]

    def get_eval_freq(self, level: int, idx: int) -> float:
        """
        Get an evaluation frequency

        Parameters
        ----------
        level : int
            The level
        idx : int
            Evaluation frequency index

        Returns
        -------
        float
            The evaluation frequency
        """
        self.check_level(level)
        self.check_eval_idx(idx)
        idx = self.per_level * level + idx
        return self.eval_freqs[idx]

    def get_fs(self, level: int) -> float:
        """
        Get sampling frequency for level

        Parameters
        ----------
        level : int
            The decimation level

        Returns
        -------
        float
            Sampling frequency Hz
        """
        self.check_level(level)
        return self.dec_fs[level]

    def get_total_factor(self, level: int) -> int:
        """
        Get total decimation factor for a level

        Parameters
        ----------
        level : int
            The level

        Returns
        -------
        int
            The decimation factor
        """
        self.check_level(level)
        return self.dec_factors[level]

    def get_incremental_factor(self, level: int) -> int:
        """
        Get incremental decimation factor

        Parameters
        ----------
        level : int
            The level

        Returns
        -------
        int
            The incremental decimation factor
        """
        self.check_level(level)
        return self.dec_increments[level]

    def to_numpy(self) -> np.ndarray:
        """Get evaluation frequencies as a 2-D array"""
        return np.array(self.eval_freqs).reshape(self.n_levels, self.per_level)

    def to_dataframe(self) -> pd.DataFrame:
        """Provide decimation parameters as DataFrame"""
        df = pd.DataFrame(data=self.to_numpy())
        df["fs"] = self.dec_fs
        df["factors"] = self.dec_factors
        df["increments"] = self.dec_increments
        df.index.name = "decimation level"
        return df


class DecimationSetup(ResisticsProcess):
    """
    Process to calculate decimation parameters

    Parameters
    ----------
    n_levels : int, optional
        Number of decimation levels, by default 8
    per_level : int, optional
        Number of frequencies per level, by default 5
    min_samples : int, optional
        Number of samples to under which to quit decimating
    div_factor : int, optional
        Minimum division factor for decimation, by default 2.
    eval_freqs : Optional[List[float]], optional
        Explicit definition of evaluation frequencies as a flat list, by
        default None. Must be of size n_levels * per_level

    Examples
    --------
    >>> from resistics.decimate import DecimationSetup
    >>> dec_setup = DecimationSetup(n_levels=3, per_level=2)
    >>> dec_params = dec_setup.run(128)
    >>> print(dec_params.to_dataframe())
                         0          1     fs  factors  increments
    decimation level
    0                 32.0  22.627417  128.0        1           1
    1                 16.0  11.313708   64.0        2           2
    2                  8.0   5.656854   32.0        4           2
    """

    n_levels: int = 8
    per_level: int = 5
    min_samples: int = 256
    div_factor: int = 2
    eval_freqs: Optional[List[float]] = None

    def run(self, fs: float) -> DecimationParameters:
        """
        Run DecimationSetup

        Parameters
        ----------
        fs : float
            Sampling frequency, Hz

        Returns
        -------
        DecimationParameters
            Decimation parameterisation
        """
        eval_freqs = self._get_eval_freqs(fs)
        factors = self._get_factors(fs, eval_freqs)
        return DecimationParameters(
            fs=fs,
            n_levels=self.n_levels,
            per_level=self.per_level,
            min_samples=self.min_samples,
            eval_freqs=eval_freqs.tolist(),
            dec_factors=factors,
        )

    def _get_eval_freqs(self, fs: float) -> np.ndarray:
        """
        Get evaluation frequencies

        Parameters
        ----------
        fs : float
            Sampling frequency, Hz

        Returns
        -------
        np.ndarray
            Evaluation frequencies as a flat array

        Raises
        ------
        ValueError
            If size of evaluation frequencies != n_levels * per_level
        ValueError
            If any evaluation frequency is > nyquist
        """
        nyquist = fs / 2
        n_freqs = self.n_levels * self.per_level
        if self.eval_freqs is None:
            return get_eval_freqs(fs, n_freqs=n_freqs)
        eval_freqs = np.array(self.eval_freqs)
        if eval_freqs.size != n_freqs:
            raise ValueError(f"Size eval freqs {eval_freqs.size} != n_freqs {n_freqs}")
        if np.any(eval_freqs > nyquist):
            raise ValueError(f"Found frequencies {eval_freqs} > nyquist {nyquist}")
        return eval_freqs

    def _get_factors(self, fs: float, eval_freqs: np.ndarray) -> List[int]:
        """
        Calculate decimation factors

        Uses evaluation frequencies, number levels and number of evaluation
        frequencies per level to calculate decimation factors with reference to
        the original sampling frequency.

        Parameters
        ----------
        fs : float
            Sampling frequency, Hz
        eval_freqs : np.ndarray
            Evaluation frequency as flat array

        Returns
        -------
        factors : List[int]
            The decimation factors referenced to the sampling frequency
        """
        eval_freqs = eval_freqs.reshape(self.n_levels, self.per_level)
        factors = []
        for ilevel in range(0, self.n_levels):
            factor = self._get_decimation_factor(fs, np.max(eval_freqs[ilevel]))
            factors.append(factor)
        return factors

    def _get_decimation_factor(self, fs: float, f_max: float) -> int:
        """
        Get decimation factor

        Given the maximum evaluation frequency in the decimation level, the
        method aims to find a suitable sampling frequency that satisfies

        - level sampling frequency >= 4 * max level evaluation frequency,

        and returns the decimation factor for this suitable sampling frequency.

        .. note::

            From an earlier check, know that all evaluation frequencies are less
            than or equal to nyquist, so this does not have to be checked again.

        Parameters
        ----------
        fs : float
            Sampling frequency, Hz
        f_max : float
            Maximum evaluation frequency in the decimation level

        Returns
        -------
        int
            The decimation factor
        """
        target_fs = f_max * 4

        if target_fs > fs:
            return 1

        factor = 1
        while True:
            test_fs = fs / self.div_factor
            if test_fs >= target_fs:
                fs = test_fs
                factor *= self.div_factor
            else:
                break
        return factor


class DecimatedLevelMetadata(Metadata):
    """Metadata for a decimation level"""

    fs: float
    """The sampling frequency of the decimation level"""
    n_samples: int
    """The number of samples in the decimation level"""
    first_time: HighResDateTime
    """The first time in the decimation level"""
    last_time: HighResDateTime
    """The last time in the decimation level"""

    @property
    def dt(self):
        return 1 / self.fs


class DecimatedMetadata(WriteableMetadata):
    """Metadata for DecimatedData"""

    fs: List[float]
    chans: List[str]
    n_chans: Optional[int] = None
    n_levels: int
    first_time: HighResDateTime
    last_time: HighResDateTime
    system: str = ""
    serial: str = ""
    wgs84_latitude: float = -999.0
    wgs84_longitude: float = -999.0
    easting: float = -999.0
    northing: float = -999.0
    elevation: float = -999.0
    chans_metadata: Dict[str, ChanMetadata]
    levels_metadata: List[DecimatedLevelMetadata]
    history: History = History()

    class Config:

        extra = "ignore"


class DecimatedData(ResisticsData):
    """
    Data class for storing decimated data

    The data for is stored in a dictionary attribute named data. The indices are
    integers representing the decimation level. Each decimation level is a
    numpy array of shape:

    n_chans x n_samples

    Parameters
    ----------
    metadata : DecimatedMetadata
        The metadata
    data : Dict[int, TimeData]
        The decimated time data

    Examples
    --------
    .. plot::
        :width: 90%

        >>> import matplotlib.pyplot as plt
        >>> from resistics.testing import time_data_random
        >>> from resistics.decimate import DecimationSetup, Decimator
        >>> time_data = time_data_random(fs=256, n_samples=10_000)
        >>> dec_params = DecimationSetup(n_levels=4, per_freq=4).run(time_data.metadata.fs)
        >>> dec_data = Decimator().run(dec_params, time_data)
        >>> for level_metadata in dec_data.metadata.levels_metadata:
        ...     level_metadata.summary()
        {
            'fs': 256.0,
            'n_samples': 10000,
            'first_time': '2020-01-01 00:00:00.000000_000000_000000_000000',
            'last_time': '2020-01-01 00:00:39.058593_750000_000000_000000'
        }
        {
            'fs': 64.0,
            'n_samples': 2500,
            'first_time': '2020-01-01 00:00:00.000000_000000_000000_000000',
            'last_time': '2020-01-01 00:00:39.046875_000000_000000_000000'
        }
        {
            'fs': 8.0,
            'n_samples': 313,
            'first_time': '2020-01-01 00:00:00.000000_000000_000000_000000',
            'last_time': '2020-01-01 00:00:39.000000_000000_000000_000000'
        }
        >>> for ilevel in range(0, dec_data.metadata.n_levels):
        ...     data = dec_data.get_level(ilevel)
        ...     plt.plot(dec_data.get_x(ilevel), data[0], label=f"Level{ilevel}") # doctest: +SKIP
        >>> plt.legend(loc=3) # doctest: +SKIP
        >>> plt.tight_layout() # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
    """

    def __init__(self, metadata: DecimatedMetadata, data: Dict[int, np.ndarray]):
        """Initialise decimated data"""
        logger.debug(f"Creating DecimatedData with data type {data[0].dtype}")
        self.metadata = metadata
        self.data = data

    def get_level(self, level: int) -> np.ndarray:
        """
        Get data for a decimation level

        Parameters
        ----------
        level : int
            The level

        Returns
        -------
        np.ndarary
            The data for the decimation level

        Raises
        ------
        ValueError
            If level > max_level
        """
        if level >= self.metadata.n_levels:
            raise ValueError(f"Level {level} not <= max {self.metadata.n_levels}")
        return self.data[level]

    def get_timestamps(
        self, level: int, samples: Optional[np.ndarray] = None, estimate: bool = True
    ) -> Union[np.ndarray, pd.DatetimeIndex]:
        """
        Get an array of timestamps

        Parameters
        ----------
        level : int
            The decimation level
        samples : Optional[np.ndarray], optional
            If provided, timestamps are only returned for the specified samples,
            by default None
        estimate : bool, optional
            Flag for using estimates instead of high precision datetimes, by
            default True

        Returns
        -------
        Union[np.ndarray, pd.DatetimeIndex]
            The return dates. This will be a numpy array of RSDateTime objects
            if estimate is False, else it will be a pandas DatetimeIndex

        Raises
        ------
        ValueError
            If the level is not within bounds
        """
        from resistics.sampling import datetime_array, datetime_array_estimate

        if level >= self.metadata.n_levels:
            raise ValueError(f"Level {level} not <= max {self.metadata.n_levels}")
        fnc = datetime_array_estimate if estimate else datetime_array
        level_metadata = self.metadata.levels_metadata[level]
        if samples is not None:
            return fnc(level_metadata.first_time, level_metadata.fs, samples=samples)
        return fnc(
            level_metadata.first_time,
            level_metadata.fs,
            n_samples=level_metadata.n_samples,
        )

    def plot(self, max_pts: int = 10_000) -> go.Figure:
        """
        Plot the decimated data

        Parameters
        ----------
        max_pts : int, optional
            Maximum number of points to plot, by default 10_000

        Returns
        -------
        go.Figure
            Plotly Figure
        """
        from resistics.time import TimeMetadata

        if len(self.data) == 0:
            logger.error("Data is empty, no decimation levels to plot")

        fig = None
        for ilevel in range(0, self.metadata.n_levels):
            logger.info(f"Plotting decimation level {ilevel}")
            metadata_dict = self.metadata.levels_metadata[ilevel].dict()
            metadata_dict["chans"] = self.metadata.chans
            metadata_dict["chans_metadata"] = self.metadata.chans_metadata
            time_data = TimeData(TimeMetadata(**metadata_dict), self.data[ilevel])
            if fig is None:
                fig = time_data.plot(
                    max_pts=max_pts,
                    label_prefix=f"Level {ilevel}",
                )
            else:
                time_data.plot(
                    fig=fig,
                    max_pts=max_pts,
                    label_prefix=f"Level {ilevel}",
                )
        return fig

    def x_size(self, level: int) -> int:
        """
        For abstract plotting functions, return the size

        Parameters
        ----------
        level : int
            The decimation level

        Returns
        -------
        int
            The x size, equal to the number of samples
        """
        return self.metadata.levels_metadata[level].n_samples

    def get_x(
        self, level: int, samples: Optional[np.ndarray] = None
    ) -> pd.DatetimeIndex:
        """
        For plotting, get x dimension, in this case times

        Parameters
        ----------
        level : int
            The decimation level
        samples : Union[np.ndarray, None], optional
            If provided, x values (timestamps) are only returned for the
            specified samples, by default None

        Returns
        -------
        pd.DatetimeIndex
            Timestamp array
        """
        return self.get_timestamps(level, samples=samples, estimate=True)

    def to_string(self) -> str:
        """Class details as a string"""
        outstr = f"{self.type_to_string()}\n"
        data = [
            [ilevel, x.fs, x.dt, x.n_samples, x.first_time, x.last_time]
            for ilevel, x in enumerate(self.metadata.levels_metadata)
        ]
        df = pd.DataFrame(
            data=data,
            columns=["level", "fs", "dt", "n_samples", "first_time", "last_time"],
        )
        df = df.set_index("level")
        outstr += df.to_string()
        return outstr


class Decimator(ResisticsProcess):
    """
    Decimate the time data into multiple levels

    There are two options for decimation, using time data Resample or using
    time data Decimate. The default is to use Resample.
    """

    resample: bool = True
    """Boolean flag for using resampling instead of decimation"""
    max_single_factor: conint(ge=3) = 3
    """Maximum single decimation factor, only used if resample is False"""

    def run(
        self, dec_params: DecimationParameters, time_data: TimeData
    ) -> DecimatedData:
        """
        Decimate the TimeData

        Parameters
        ----------
        dec_params : DecimationParameters
            The decimation parameters
        time_data : TimeData
            TimeData to decimate

        Returns
        -------
        DecimatedData
            DecimatedData instance with all the decimated data
        """
        decimation_fnc = self._resample if self.resample else self._decimate
        metadata_dict = time_data.metadata.dict()
        data = {}
        levels_metadata = []
        messages = []
        for ilevel in range(0, dec_params.n_levels):
            factor = dec_params.dec_increments[ilevel]
            logger.info(f"Decimating level {ilevel} with factor {factor}")
            time_data_new = decimation_fnc(time_data, factor)
            if time_data_new.metadata.n_samples < dec_params.min_samples:
                logger.warning(f"n_samples < min allowed {dec_params.min_samples}")
                break
            data[ilevel] = time_data_new.data
            time_data = time_data_new
            fs = dec_params.dec_fs[ilevel]
            levels_metadata.append(DecimatedLevelMetadata(**time_data.metadata.dict()))
            messages.append(f"Decimated level {ilevel}, inc. factor {factor}, fs {fs}")
        completed = list(range(len(data)))
        target = list(range(dec_params.n_levels))
        messages.append(f"Completed levels {completed} out of {target}")
        metadata = self._get_metadata(metadata_dict, levels_metadata)
        metadata.history.add_record(self._get_record(messages))
        return DecimatedData(metadata, data)

    def _decimate(self, time_data: TimeData, factor: int) -> TimeData:
        """Decimate time data using decimate"""
        from resistics.time import Decimate

        if factor == 1:
            return time_data
        decimator = Decimate(factor=factor, max_single_factor=self.max_single_factor)
        return decimator.run(time_data)

    def _resample(self, time_data: TimeData, factor: int) -> TimeData:
        """Decimate time data using resampling"""
        from resistics.time import Resample

        if factor == 1:
            return time_data
        return Resample(new_fs=time_data.metadata.fs / factor).run(time_data)

    def _get_metadata(
        self,
        metadata_dict: Dict[str, Any],
        levels_metadata: List[DecimatedLevelMetadata],
    ) -> DecimatedMetadata:
        """Get the metadata for the decimated data"""
        metadata_dict["fs"] = [x.fs for x in levels_metadata]
        metadata_dict["n_levels"] = len(levels_metadata)
        metadata_dict["levels_metadata"] = levels_metadata
        return DecimatedMetadata(**metadata_dict)


class DecimatedDataWriter(ResisticsWriter):
    """Writer of resistics decimated data"""

    def run(self, dir_path: Path, dec_data: DecimatedData) -> None:
        """
        Write out DecimatedData

        Parameters
        ----------
        dir_path : Path
            The directory path to write to
        dec_data : DecimatedData
            Decimated data to write out

        Raises
        ------
        WriteError
            If unable to write to the directory
        """
        from resistics.errors import WriteError

        if not self._check_dir(dir_path):
            raise WriteError(dir_path, "Unable to write to directory, check logs")
        logger.info(f"Writing decimated data to {dir_path}")
        metadata_path = dir_path / "metadata.json"
        data_path = dir_path / "data"
        np.savez_compressed(data_path, **{str(x): y for x, y in dec_data.data.items()})
        metadata = dec_data.metadata.copy()
        metadata.history.add_record(self._get_record(dir_path, type(dec_data)))
        metadata.write(metadata_path)


class DecimatedDataReader(ResisticsProcess):
    """Reader of resistics decimated data"""

    def run(
        self, dir_path: Path, metadata_only: bool = False
    ) -> Union[DecimatedMetadata, DecimatedData]:
        """
        Read DecimatedData

        Parameters
        ----------
        dir_path : Path
            The directory path to read from
        metadata_only : bool, optional
            Flag for getting metadata only, by default False

        Returns
        -------
        Union[DecimatedMetadata, DecimatedData]
            DecimatedData or DecimatedMetadata if metadata_only

        Raises
        ------
        ReadError
            If the directory does not exist
        """
        from resistics.errors import ReadError

        if not dir_path.exists():
            raise ReadError(dir_path, "Directory does not exist")
        logger.info(f"Reading decimated data from {dir_path}")
        metadata_path = dir_path / "metadata.json"
        metadata = DecimatedMetadata.parse_file(metadata_path)
        if metadata_only:
            return metadata
        data_path = dir_path / "data.npz"
        npz_file = np.load(data_path)
        data = {int(level): npz_file[level] for level in npz_file.files}
        messages = [f"Decimated data read from {dir_path}"]
        metadata.history.add_record(self._get_record(messages))
        return DecimatedData(metadata, data)
